# -*- coding: utf-8 -*-
"""
This module provides a currency table class that a stores a
dictionary of currency with exchange rates based on the table's default currency.\n

@name: currency table
@author: Ex Mente Technologies (Pty) Ltd
"""

from auxi.core.namedobject import NamedObject
from auxi.modeling.financial.des.currency import Currency

__version__ = "0.2.0"

# TODO: initialise accounts by json file
# TODO: get_account by path.


class CurrencyTable(NamedObject):
    """Represents a currency table that is used to define the exchange rates of
    different currencies based on the default currency."""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, description=None,
                 default_currency=Currency(
                    "USD", "United States Dollars", "$")):
        """Initialise the object.

        :param name: The name.
        :param description: The description.
        """
        super().__init__(name, description)
        self.default_currency = default_currency
        self.table = {}
        self.table[self.default_currency] = "1.0"



# -*- coding: utf-8 -*-
"""
Created on Mon Jan 18 05:34:00 2016

@author: johanz
"""

import json
import jsonpickle


class Object(object):
    """
    Base class for all auxi classes.
    """

    def __str__(self):
        o = json.loads(jsonpickle.encode(self))
        result = json.dumps(o, sort_keys=True,
                            indent=4, separators=(',', ': '))
        return result

    def __hash__(self):
        return hash(str(self))

if __name__ == "__main__":
    o = Object()
    print(o)
    print(o.__hash__())

    str_o = str(o)
    new_o = jsonpickle.decode(str_o)
    print(new_o)
    print(type(new_o))



# -*- coding: utf-8 -*-
"""
This module provides testing code for the
auxi.modeling.business.component module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
from datetime import datetime
from auxi.modeling.business.clock import Clock
from auxi.modeling.business.component import Component
from auxi.modeling.business.basicactivity import BasicActivity
from auxi.modeling.financial.des.generalledger import GeneralLedger
from auxi.modeling.financial.des.generalledgerstructure import GeneralLedgerStructure
from auxi.modeling.financial.des.transactiontemplate import TransactionTemplate

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class TestAllFunctions(unittest.TestCase):
    """
      Tester for the auxi.modeling.business.component class.
    """

    def setUp(self):
        self.object = Component("ComponentA",
                                description="DescriptionA")
        # Set up the needed objects
        self.object.components.append(
            Component("ComponentA1", description="ca1"))
        self.object.components.append(
            Component("ComponentA2", description="ca2"))
        self.tx_template = TransactionTemplate("NameA",
                                               description="DescriptionA",
                                               dt_account="Bank",
                                               cr_account="Sales")
        basic_activity = BasicActivity("BasicActivityA",
                                       description="DescriptionA",
                                       start=datetime(2016, 2, 1),
                                       end=datetime(2017, 2, 1),
                                       interval=3,
                                       amount=5000,
                                       tx_template=self.tx_template)
        self.object.components[0].activities.append(basic_activity)

    def test_constructor(self):
        self.assertEqual(self.object.name, "ComponentA")
        self.assertEqual(self.object.description, "DescriptionA")

    def test_set_parent_path(self):
        self.object.set_parent_path("entityA")
        self.assertEqual(self.object.path, "entityA/ComponentA")

    def test_set_name(self):
        self.object.set_parent_path("entityA")
        self.object.name = "NameAt"
        self.assertEqual(self.object.name, "NameAt")
        self.assertEqual(self.object.components[0].path,
                         "entityA/NameAt/ComponentA1")
        self.assertEqual(self.object.components[1].path,
                         "entityA/NameAt/ComponentA2")
        self.assertEqual(self.object.components[0].activities[0].path,
                         "entityA/NameAt/ComponentA1/BasicActivityA")

    def test_create_component(self):
        new_comp = self.object.create_component("ComponentTestCreate",
                                                description="catc")
        self.assertEqual(new_comp.name, "ComponentTestCreate")
        self.assertEqual(new_comp.description, "catc")

        self.assertEqual(new_comp, self.object.components[2])

    def test_remove_component(self):
        self.object.create_component("ComponentTestRemove",
                                     description="catr")
        self.object.remove_component("ComponentTestRemove")
        self.assertEqual(len(self.object.components), 2)

    def test_prepare_to_run(self):
        clock = Clock("NameA", start_datetime=datetime(2016, 1, 1))
        self.object.prepare_to_run(clock, 18)
        self.assertEqual(
            self.object.components[0].activities[0].start_period_ix, 2)
        self.assertEqual(
            self.object.components[0].activities[0].end_period_ix, 14)

    def test_run(self):
        structure = GeneralLedgerStructure("NameA", description="DescriptionA")
        gl = GeneralLedger("NameA", structure, description="DescriptionA")
        clock = Clock("NameA", start_datetime=datetime(2016, 1, 1))
        clock.tick()
        clock.tick()
        self.object.prepare_to_run(clock, 20)
        self.object.run(clock, gl)
        self.assertEqual(len(gl.transactions), 1)
        self.assertEqual(gl.transactions[0].amount, 5000)

# =============================================================================
# Display documentation and run tests.
# =============================================================================
# os.system("cls")

# help(Component)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
This module contains a single class that represents chemical compound.

@author: Ex Mente Technologies (Pty) Ltd
"""

from auxi.core.object import Object
from auxi.tools.chemistry import stoichiometry as stoich
from auxi.tools.chemistry.phase import Phase

__version__ = "0.2.0"


class Compound(Object):
    """Represents a chemical compound.

    :param dictionary: Dictionary containing the data required to initialise \
    the compound."""

    def __init__(self, dictionary):
        self.formula = dictionary["Formula"]
        """Chemical formula, e.g. "Fe", "CO2"."""

        self.molar_mass = stoich.molar_mass(self.formula) / 1000.0
        """Molar mass. [kg/mol]"""

        self._phases = dict()
        """Dictionary containing the compound's phase objects."""

        for k, v in dictionary["Phases"].items():
            self._phases[k] = Phase(v)

    def __str__(self):
        result = "COMPOUND: " + "\n"
        result = result + "\tFormula: " + self.formula + "\n"
        for k, v in self._phases.items():
            result = result + str(v)
        return result

    def get_phase_list(self):
        """Get a list of the compound's phases.

        :returns: List of phases."""

        return sorted(self._phases.keys())

    def Cp(self, phase, temperature):
        """Calculate the heat capacity of a phase of the compound at a
        specified temperature.

        :param phase:       A phase of the compound, e.g. "S", "L", "G".
        :param temperature: [K]

        :returns: Heat capacity. [J/mol/K]"""

        try:
            return self._phases[phase].Cp(temperature)
        except KeyError:
            raise Exception("The phase '" + phase + "' was not found in compound '" + self.formula + "'.")

    def H(self, phase, temperature):
        """Calculate the enthalpy of a phase of the compound at a specified
        temperature.

        :param phase:       A phase of the compound, e.g. "S", "L", "G".
        :param temperature: [K]

        :returns: Enthalpy. [J/mol]"""

        try:
            return self._phases[phase].H(temperature)
        except KeyError:
            raise Exception("The phase '" + phase + "' was not found in compound '" + self.formula + "'.")

    def S(self, phase, temperature):
        """Calculate the enthalpy of a phase of the compound at a specified
        temperature.

        :param phase:       A phase of the compound, e.g. "S", "L", "G".
        :param temperature: [K]

        :returns: Entropy. [J/mol/K]"""

        try:
            return self._phases[phase].S(temperature)
        except KeyError:
            raise Exception("The phase '" + phase + "' was not found in compound '" + self.formula + "'.")

    def G(self, phase, temperature):
        """Calculate the Gibbs free energy of a phase of the compound at a
        specified temperature.

        :param phase:       A phase of the compound, e.g. "S", "L", "G".
        :param temperature: [K]

        :returns: Gibbs free energy. [J/mol]"""

        try:
            return self._phases[phase].G(temperature)
        except KeyError:
            raise Exception("The phase '" + phase + "' was not found in compound '" + self.formula + "'.")



# -*- coding: utf-8 -*-
"""
This module provides an activity class that serves as a base class
for business activities.\n

@name: activity
@author: Ex Mente Technologies (Pty) Ltd
"""

from datetime import datetime
from auxi.core.namedobject import NamedObject

__version__ = "0.2.0"

# TODO: initialise accounts by json file
# TODO: get_account by path.


class Activity(NamedObject):
    """Represents an activity base class."""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, description=None,
                 start=datetime.min,
                 end=datetime.max,
                 interval=1):
        """Initialise the object.

        :param name: The name.
        :param description: The description.
        :param start: The datetime the activity should be started.
        :param end: The datetime the activity should be run until.
        :param interval: The interval of the activity.
        """
        self._parent_path = ""
        super().__init__(name, description)
        self.start_datetime = start
        self.end_datetime = end
        self.interval = interval
        self.start_period_ix = -1
        self.end_period_ix = -1
        self.period_count = -1

    def set_parent_path(self, value):
        self._parent_path = value
        self.path = value + r'/' + self.name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self.path = self._parent_path + r'/' + self.name

    def _meet_execution_criteria(self, ix_period):
        if self.interval != 0 and (ix_period+1) % self.interval != 0:
            return False
        return ix_period >= self.start_period_ix and ix_period + self.interval <= self.end_period_ix

    def prepare_to_run(self, clock, period_count):
        """Prepare the activity for execution.

        :param clock: The clock containing the execution start time and execution period information.
        :param period_count: The total amount of periods this activity will be requested to be run for.
        """
        if self.start_period_ix == -1 and self.start_datetime != datetime.min:
            # Set the Start period index
            for i in range(0, period_count):
                if clock.get_datetime_at_period_ix(i) > self.start_datetime:
                    self.start_period_ix = i
                    break
        if self.start_period_ix == -1:
            self.start_period_ix = 0
        if self.period_count == -1 and self.end_datetime != datetime.max:
            # Set the Start date
            for i in range(0, period_count):
                if clock.get_datetime_at_period_ix(i) > self.end_datetime:
                    self.period_count = i - self.start_period_ix
                    break
        if self.period_count != -1:
            self.end_period_ix = self.start_period_ix + self.period_count
        else:
            self.end_period_ix = self.start_period_ix + period_count



# -*- coding: utf-8 -*-
"""
This module provides a general ledger class that a stores a
structure of accounts and list of transactions between these accounts.\n

@name: general ledger
@author: Ex Mente Technologies (Pty) Ltd
"""

from datetime import datetime
from auxi.core.namedobject import NamedObject
from auxi.modeling.financial.des.transaction import Transaction

__version__ = "0.2.0"

# TODO: initialise accounts by json file
# TODO: get_account by path.


class GeneralLedger(NamedObject):
    """Represents the account structure of a general ledger."""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, structure, description=None):
        """Initialise the object.

        :param name: The name.
        :param structure: The general ledger structure.
        :param description: The description.
        """
        super().__init__(name, description)
        self.structure = structure
        self.transactions = []

    def create_transaction(self, name, description=None,
                           tx_datetime=datetime.min,
                           dt_account=None, cr_account=None,
                           source=None, amount=0.00):
        """Create a transaction in the general ledger.

        :param name: The transaction's name.
        :param description: The transaction's description.
        :param tx_datetime: The date and time of the transaction.
        :param cr_account: The transaction's credit account's name.
        :param dt_account: The transaction's debit account's name.
        :param source: The name of source the transaction originated from.
        :param amount: The transaction amount.

        :returns: The created transaction.
        """
        new_tx = Transaction(name, description, tx_datetime,
                             dt_account, cr_account, source, amount)
        self.transactions.append(new_tx)
        return new_tx



# -*- coding: utf-8 -*-
"""
This module contains a single class representing a phase of a chemical
compound.

@author: Ex Mente Technologies (Pty) Ltd
"""

import math
from auxi.core.namedobject import NamedObject
from auxi.tools.chemistry.cprecord import CpRecord

__version__ = "0.2.0"


class Phase(NamedObject):
    """Represents a phase of a chemical compound.

    :param dictionary: Dictionary containing the data required to initialise \
    the phase."""

    def __init__(self, dictionary):
        self.name = dictionary["Symbol"]
        """The phase's name, e.g. solid, liquid, gas, etc."""

        self.symbol = dictionary["Symbol"]
        """The phase's symbol, e.g. S1 = solid 1, L = liquid, etc."""

        self.Tref = 298.15
        """The reference temperature of the phase. [K]"""

        self.DHref = dictionary["DHref"]
        """The formation enthalpy of the phase at Tref. [J/mol]"""

        self.Sref = dictionary["Sref"]
        """The standard entropy of the phase at Tref. [J/mol/K]"""

        self._Cp_records = dict()
        """A dictionary containing the phase's Cp records."""

        for k, v in dictionary["Cp_records"].items():
            self._Cp_records[k] = CpRecord(v)

    def __str__(self):
        result = "\tPHASE: " + self.name + "\n"
        result = result + "\t\tName: " + self.name + "\n"
        result = result + "\t\tSymbol: " + self.symbol + "\n"
        result = result + "\t\tTref: " + str(self.Tref) + "\n"
        result = result + "\t\tDHref: " + str(self.DHref) + "\n"
        result = result + "\t\tSref: " + str(self.Sref) + "\n"
        result = result + "\t\tCp record count:" + str(len(self._Cp_records)) + "\n"
        for k, v in self._Cp_records.items():
            result = result + str(v)
        return result

    def Cp(self, temperature):
        """Calculate the heat capacity of the compound phase at the specified
        temperature.

        :param temperature: [K]

        :returns: The heat capacity of the compound phase. [J/mol/K]
        """
        for Tmax in sorted(self._Cp_records.keys()):
            if temperature < Tmax:
                return self._Cp_records[Tmax].Cp(temperature)
        Tmax = max(self._Cp_records.keys())
        return self._Cp_records[Tmax].Cp(Tmax)

    def H(self, temperature):
        """Calculate the enthalpy of the compound phase at the specified
        temperature.

        :param temperature: [K]

        :returns: The enthalpy of the compound phase. [J/mol]
        """
        result = self.DHref
        for Tmax in sorted(self._Cp_records.keys()):
            result = result + self._Cp_records[Tmax].H(temperature)
            if temperature <= Tmax:
                return result
        # Extrapolate beyond the upper limit by using a constant heat capacity.
        Tmax = max(self._Cp_records.keys())
        result = result + self.Cp(Tmax)*(temperature - Tmax)
        return result

    def S(self, temperature):
        """Calculate the entropy of the compound phase at the specified
        temperature.

        :param temperature: [K]

        :returns: The entropy of the compound phase. [J/mol/K]
        """
        result = self.Sref
        for Tmax in sorted(self._Cp_records.keys()):
            result = result + self._Cp_records[Tmax].S(temperature)
            if temperature <= Tmax:
                return result
        # Extrapolate beyond the upper limit by using a constant heat capacity.
        Tmax = max(self._Cp_records.keys())
        result = result + self.Cp(Tmax)*math.log(temperature / Tmax)
        return result

    def G(self, temperature):
        """Calculate the heat capacity of the compound phase at the specified
        temperature.

        :param temperature: [K]

        :returns: The Gibbs free energy of the compound phase. [J/mol]
        """
        h = self.DHref
        s = self.Sref
        for Tmax in sorted(self._Cp_records.keys()):
            h = h + self._Cp_records[Tmax].H(temperature)
            s = s + self._Cp_records[Tmax].S(temperature)
            if temperature <= Tmax:
                return h - temperature * s
        # Extrapolate beyond the upper limit by using a constant heat capacity.
        Tmax = max(self._Cp_records.keys())
        h = h + self.Cp(Tmax)*(temperature - Tmax)
        s = s + self.Cp(Tmax)*math.log(temperature / Tmax)
        return h - temperature * s



# -*- coding: utf-8 -*-
"""
This module provides a basic activity class that provides the most basic functionality:
periodically create a transaction between two specified accounts.\n

@name: basic activity
@author: Ex Mente Technologies (Pty) Ltd
"""

from datetime import datetime
from auxi.core.namedobject import NamedObject
from auxi.modeling.financial.des.transactiontemplate import TransactionTemplate
from auxi.modeling.business.activity import Activity

__version__ = "0.2.0"

# TODO: initialise accounts by json file
# TODO: get_account by path.


class BasicActivity(Activity):
    """Represents an basic activity class."""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, description=None,
                 start=datetime.min,
                 end=datetime.max,
                 interval=1,
                 amount=1000,
                 tx_template=TransactionTemplate("Unknown")):
        """Initialise the object.

        :param name: The name.
        :param description: The description.
        :param start: The datetime the activity should be started.
        :param end: The datetime the activity should be run until.
        :param interval: The interval of the activity.
        :param amount: The amount of an activity.
        :param tx_template: The template for the transaction.
        """
        super().__init__(name,
                         description=description,
                         start=start,
                         end=end,
                         interval=interval)
        self.start_datetime = start
        self.end_datetime = end
        self.interval = interval
        self.amount = amount
        self.tx_template = tx_template

    def _meet_execution_criteria(self, ix_period):
        return super()._meet_execution_criteria(ix_period) and self.amount > 0

    def run(self, clock, generalLedger):
        """Execute the activity at the current clock cycle.

        :param clock: The clock containing the current execution time and period information.
        :param generalLedger: The general ledger into which to create the transactions.
        """
        if not self._meet_execution_criteria(clock.timestep_ix):
            return

        generalLedger.create_transaction(
            self.tx_template.name,
            description=self.tx_template.description,
            tx_datetime=clock.get_datetime(),
            dt_account=self.tx_template.dt_account,
            cr_account=self.tx_template.cr_account,
            source=self.path,
            amount=self.amount)



# -*- coding: utf-8 -*-
"""
This module provides a sales tax rule class for specifying a rule that
handle sales taxes.\n

@name: sales tax rule
@author: Ex Mente Technologies (Pty) Ltd
"""

from auxi.core.namedobject import NamedObject

__version__ = "0.2.0"

# TODO: initialise accounts by json file
# TODO: get_account by path.


class SalesRule(NamedObject):
    """Represents a sales tax rule class."""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, description=None, percentage=1.0):
        """Initialise the object.

        :param name: The name.
        :param description: The description.
        :param percentage: The sales tax percentage.
        """
        super().__init__(name, description)
        self.percentage = percentage



# -*- coding: utf-8 -*-
"""
This module contains a single class that represents an element of the periodic
table.

@author: Ex Mente Technologies (Pty) Ltd
"""

from auxi.core.object import Object

__version__ = "0.2.0"


class Element(Object):
    """An element in the periodic table.

    :param period:        Period to which the element belongs.
    :param group:         Group to which the element belongs.
    :param atomic_number: Number of protons in the element's nucleus.
    :param symbol:        Element's symbol.
    :param molar_mass:    [kg/kmol] Element's standard atomic mass.
    """

    def __init__(self, period, group, atomic_number, symbol, molar_mass):
        self.period = period
        self.group = group
        self.atomic_number = atomic_number
        self.symbol = symbol
        self.molar_mass = molar_mass



# -*- coding: utf-8 -*-
"""
This module provides a time based model class that represents a
time based business model.
A time based model contains and executes entities.\n

@name: time based model
@author: Ex Mente Technologies (Pty) Ltd
"""

from datetime import datetime
from auxi.core.namedobject import NamedObject
from auxi.modeling.business.clock import Clock
from auxi.modeling.business.clock import TimePeriod
from auxi.modeling.business.entity import Entity
from auxi.modeling.financial.des.generalledgerstructure import GeneralLedgerStructure

__version__ = "0.2.0"

# TODO: initialise accounts by json file
# TODO: get_account by path.


class TimeBasedModel(NamedObject):
    """Represents an time based model class."""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self,
                 name,
                 description=None,
                 start_datetime=datetime.now(),
                 period_duration=TimePeriod.month,
                 period_count=60):
        """Initialise the object.

        :param name: The name.
        :param description: The description.
        :param start_datetime: The start datetime of the model.
        :param period_duration: The duration of the model's time period. e.g. month, day etc.
        :param start_datetime: The number of periods to execute the model for.
        """
        self.clock = Clock(
            "Clock",
            start_datetime=start_datetime,
            timestep_period_duration=period_duration)
        self.period_count = period_count
        self.entities = []
        super().__init__(name, description=description)

    def _update_childrens_parent_path(self):
        for e in self.entities:
            e.set_parent_path(self.name)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self._update_childrens_parent_path()

    def create_entity(self, name, gl_structure, description=None):
        """Create an entity in the model.

        :param name: The entity name.
        :param gl_structure: The entity's general ledger structure.
        :param description: The entity description.

        :returns: The created entity.
        """
        new_entity = Entity(name, gl_structure, description=description)
        self.entities.append(new_entity)
        return new_entity

    def remove_entity(self, name):
        """Remove an entity from the model.

        :param name: The name of the entity to remove.
        """
        entity_to_remove = None
        for e in self.entities:
            if e.name == name:
                entity_to_remove = e
        if entity_to_remove is not None:
            self.entities.remove(entity_to_remove)

    def prepare_to_run(self):
        """Prepare the model for execution.
        """
        self.clock.reset()
        for e in self.entities:
            e.prepare_to_run(self.clock, self.period_count)

    def run(self):
        """Execute the model.
        """
        self.prepare_to_run()
        for i in range(0, self.period_count):
            self.clock.tick()
            for e in self.entities:
                e.run(self.clock)



# -*- coding: utf-8 -*-
"""
This module provides a currency class representing a currency.\n

@name: currency
@author: Ex Mente Technologies (Pty) Ltd
"""

from auxi.core.namedobject import NamedObject

__version__ = "0.2.0"

# TODO: initialise accounts by json file
# TODO: get_account by path.


class Currency(NamedObject):
    """Represents a currency."""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, description=None, symbol=None):
        """Initialise the object.

        :param name: The name.
        :param description: The description.
        :param symbol: The symbol.
        """
        super().__init__(name, description)
        self.symbol = symbol



# -*- coding: utf-8 -*-
"""
This module provides testing code for the named object module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
import jsonpickle
from auxi.core.namedobject import NamedObject

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class TestAllFunctions(unittest.TestCase):
    """Tester for the auxi.core.Object class."""

    def setUp(self):
        self.object = NamedObject("NameA", "DescriptionA")

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")

    def test__str__(self):
        obj = NamedObject("NameB", "DescriptionB")
        self.assertNotEqual(str(obj), str(self.object))

        str_o = str(self.object)
        new_o = jsonpickle.decode(str_o)
        self.assertEqual(str_o, str(new_o))


# =============================================================================
# Display documentation and run tests.
# =============================================================================
# os.system("cls")

# help(Object)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
This module provides a clock class that provides functions to manage a
ticking clock based on a time period as well as retreive the current tick's
date since the start date.\n

@name: clock
@author: Ex Mente Technologies (Pty) Ltd
"""

from enum import Enum
from datetime import datetime
from datetime import timedelta
from auxi.core.namedobject import NamedObject
from dateutil.relativedelta import relativedelta

__version__ = "0.2.0"


class TimePeriod(Enum):
    millisecond = 1
    second = 2
    minute = 3
    hour = 4
    day = 5
    week = 6
    month = 7
    year = 8


class Clock(NamedObject):
    """Represents a clock."""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, description=None,
                 start_datetime=datetime.min,
                 timestep_period_duration=TimePeriod.month,
                 timestep_period_count=1):
        """Initialise the object.

        :param name: The name.
        :param description: The description.
        :param start_datetime: The start datetime.
        :param timestep_period_duration: The duration of each time period.
        :param timestep_period_count: The number of periods that makes up a timestep.
        """
        super().__init__(name, description)
        self.start_datetime = start_datetime
        self.timestep_period_duration = timestep_period_duration
        self.timestep_period_count = timestep_period_count
        self.timestep_ix = 0

    def tick(self):
        self.timestep_ix += 1

    def reset(self):
        self.timestep_ix = 0

    def get_datetime_at_period_ix(self, ix):
        if self.timestep_period_duration == TimePeriod.millisecond:
            return self.start_datetime + timedelta(milliseconds=ix)
        elif self.timestep_period_duration == TimePeriod.second:
            return self.start_datetime + timedelta(seconds=ix)
        elif self.timestep_period_duration == TimePeriod.minute:
            return self.start_datetime + timedelta(minutes=ix)
        elif self.timestep_period_duration == TimePeriod.hour:
            return self.start_datetime + timedelta(hours=ix)
        elif self.timestep_period_duration == TimePeriod.day:
            return self.start_datetime + relativedelta(days=ix)
        elif self.timestep_period_duration == TimePeriod.week:
            return self.start_datetime + relativedelta(days=ix*7)
        elif self.timestep_period_duration == TimePeriod.month:
            return self.start_datetime + relativedelta(months=ix)
        elif self.timestep_period_duration == TimePeriod.year:
            return self.start_datetime + relativedelta(years=ix)

    def get_datetime(self):
        return self.get_datetime_at_period_ix(self.timestep_ix)



# -*- coding: utf-8 -*-
"""
This module provides testing code for the
auxi.modeling.business.timebasedmodel module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
from datetime import datetime
from auxi.modeling.business.timebasedmodel import TimeBasedModel
from auxi.modeling.business.basicactivity import BasicActivity
from auxi.modeling.business.clock import TimePeriod
from auxi.modeling.financial.des.generalledgerstructure import GeneralLedgerStructure
from auxi.modeling.financial.des.transactiontemplate import TransactionTemplate

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class TestAllFunctions(unittest.TestCase):
    """
      Tester for the auxi.modeling.business.timebasedmodel class.
    """

    def setUp(self):
        self.object = TimeBasedModel("TimeBasedModelA",
                                     description="TimeBasedModel A",
                                     start_datetime=datetime(2016, 2, 1),
                                     period_duration=TimePeriod.day,
                                     period_count=200)
        # Set up the needed objects
        self.gl_structure = GeneralLedgerStructure(
            "NameA",
            description="DescriptionA")
        entity = self.object.create_entity(
            "EntityA",
            self.gl_structure,
            description="DescriptionA")
        # Set up the needed objects
        comp1 = entity.create_component("ComponentA1", description="ca1")
        tx_template = TransactionTemplate("NameA",
                                          description="DescriptionA",
                                          dt_account="Bank",
                                          cr_account="Sales")
        basic_activity = BasicActivity("BasicActivityA",
                                       description="DescriptionA",
                                       start=datetime(2016, 2, 1),
                                       end=datetime(2016, 2, 14),
                                       interval=1,
                                       amount=5000,
                                       tx_template=tx_template)
        comp1.activities.append(basic_activity)

    def test_constructor(self):
        self.assertEqual(self.object.name, "TimeBasedModelA")
        self.assertEqual(self.object.description, "TimeBasedModel A")
        self.assertEqual(self.object.clock.start_datetime,
                         datetime(2016, 2, 1))
        self.assertEqual(self.object.clock.timestep_period_duration,
                         TimePeriod.day)
        self.assertEqual(self.object.period_count, 200)

    def test_set_name(self):
        self.object.name = "NameAt"
        self.assertEqual(self.object.name, "NameAt")
        self.assertEqual(
            self.object.entities[0].components[0].path,
            "NameAt/EntityA/ComponentA1")
        self.assertEqual(
            self.object.entities[0].components[0].activities[0].path,
            "NameAt/EntityA/ComponentA1/BasicActivityA")

    def test_create_entity(self):
        new_entity = self.object.create_entity(
            "EntityTestCreate",
            self.gl_structure,
            description="ec")
        self.assertEqual(new_entity.name, "EntityTestCreate")
        self.assertEqual(new_entity.description, "ec")

        self.assertEqual(new_entity, self.object.entities[1])

    def test_remove_entity(self):
        self.object.create_entity(
            "EntityTestRemove",
            self.gl_structure,
            description="er")
        self.object.remove_entity("EntityTestRemove")
        self.assertEqual(len(self.object.entities), 1)

    def test_prepare_to_run(self):
        self.object.prepare_to_run()
        self.assertEqual(
            self.object.entities[0].components[0].activities[0].end_period_ix,
            14)

    def test_run(self):
        self.object.run()
        self.assertEqual(len(self.object.entities[0].gl.transactions), 13)
        self.assertEqual(self.object.entities[0].gl.transactions[0].amount,
                         5000)

# =============================================================================
# Display documentation and run tests.
# =============================================================================
# os.system("cls")

# help(Component)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
This module provides testing code for the
auxi.modeling.business.clock module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
from datetime import datetime
from auxi.modeling.business.clock import TimePeriod
from auxi.modeling.business.clock import Clock

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class TestAllFunctions(unittest.TestCase):
    """
      Tester for the auxi.modeling.business.clock class.
    """

    def setUp(self):
        self.object = Clock("NameA",
                            description="DescriptionA",
                            start_datetime=datetime(2016, 2, 1),
                            timestep_period_duration=TimePeriod.year,
                            timestep_period_count=3)

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.start_datetime, datetime(2016, 2, 1))
        self.assertEqual(self.object.timestep_period_duration, TimePeriod.year)
        self.assertEqual(self.object.timestep_period_count, 3)

    def test_tick(self):
        self.assertEqual(self.object.timestep_ix, 0)
        self.object.tick()
        self.assertEqual(self.object.timestep_ix, 1)

    def test_reset(self):
        self.object.tick()
        self.object.reset()
        self.assertEqual(self.object.timestep_ix, 0)

    def test_get_datetime_at_period_ix(self):
        self.assertEqual(self.object.get_datetime_at_period_ix(3),
                         datetime(2019, 2, 1))

    def test_get_datetime(self):
        self.object.tick()
        self.assertEqual(self.object.get_datetime(), datetime(2017, 2, 1))

# =============================================================================
# Display documentation and run tests.
# =============================================================================
# os.system("cls")

# help(SalesRule)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
This module provides a tax rule class that serves as a base class
for tax rules.\n

@name: tax rule
@author: Ex Mente Technologies (Pty) Ltd
"""

from auxi.core.namedobject import NamedObject

__version__ = "0.2.0"

# TODO: initialise accounts by json file
# TODO: get_account by path.


class Rule(NamedObject):
    """Represents a tax rule base class."""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, description=None):
        """Initialise the object.

        :param name: The name.
        :param description: The description.
        """
        super().__init__(name, description)



# -*- coding: utf-8 -*-
"""
This module provides a general ledger account class and an account type enum.\n

@name: general ledger account
@author: Ex Mente Technologies (Pty) Ltd
"""

from enum import Enum
from auxi.core.namedobject import NamedObject


__version__ = "0.2.0"


class AccountType(Enum):
    asset = 1
    equity = 2
    expense = 3
    liability = 4
    revenue = 5


class GeneralLedgerAccount(NamedObject):
    """Represents an account of a general ledger.
    """

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, description=None,
                 number=None,
                 account_type=AccountType.revenue):
        """Initialise the object."""
        super().__init__(name, description)
        self.number = number
        self.account_type = account_type
        self.accounts = []

    def create_account(self, name, description=None, number=None):
        """Create a sub account in the account.

        :param name: The account name.
        :param description: The account description.
        :param number: The account number.

        :returns: The created account.
        """
        new_account = GeneralLedgerAccount(name, description, number,
                                           self.account_type)
        self.accounts.append(new_account)
        return new_account

    def remove_account(self, name):
        """Remove an account from the account's sub accounts.

        :param name: The name of the account to remove.
        """
        acc_to_remove = None
        for a in self.accounts:
            if a.name == name:
                acc_to_remove = a
        if acc_to_remove is not None:
            self.accounts.remove(acc_to_remove)



# -*- coding: utf-8 -*-
"""
This module provides a tax rule set class that groups a collection of tax rules.
This is useful for distinguising between different countries' tax rules.\n

@name: tax rule set
@author: Ex Mente Technologies (Pty) Ltd
"""

from auxi.core.namedobject import NamedObject

__version__ = "0.2.0"

# TODO: initialise accounts by json file
# TODO: get_account by path.


class RuleSet(NamedObject):
    """Represents a tax rule set base class."""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, description=None, code=None):
        """Initialise the object.

        :param name: The name.
        :param description: The description.
        :param code: The code identifying the rule set. E.g. ZA tax rules, or US tax rules.
        """
        super().__init__(name, description)
        self.code = code
        self.rules = []



# -*- coding: utf-8 -*-
"""
This module provides testing code for the auxi.modeling.financial.des.transactiontemplate module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
from auxi.modeling.financial.des.transactiontemplate import TransactionTemplate

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class TestAllFunctions(unittest.TestCase):
    """Tester for the auxi.modeling.financial.des.transactiontemplate class."""

    def setUp(self):
        self.object = TransactionTemplate("NameA",
                                          description="DescriptionA",
                                          dt_account="Bank",
                                          cr_account="Sales")

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.dt_account, "Bank")
        self.assertEqual(self.object.cr_account, "Sales")


# =============================================================================
# Display documentation and run tests.
# =============================================================================
# os.system("cls")

# help(TransactionTemplate)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
This module provides testing code for the
auxi.modeling.financial.des.generalledgerstructure module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
from auxi.modeling.financial.des.generalledgeraccount import AccountType
from auxi.modeling.financial.des.generalledgerstructure import GeneralLedgerStructure

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class TestAllFunctions(unittest.TestCase):
    """
      Tester for the auxi.modeling.financial.des.generalledgerstructure class.
    """

    def setUp(self):
        self.object = GeneralLedgerStructure("NameA",
                                             description="DescriptionA")

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.accounts[0].name, "Bank")
        self.assertEqual(self.object.accounts[0].account_type,
                         AccountType.asset)
        self.assertEqual(self.object.accounts[1].name, "IncomeTaxPayable")
        self.assertEqual(self.object.accounts[1],
                         self.object.incometaxpayable_account)
        self.assertEqual(self.object.accounts[1].account_type,
                         AccountType.liability)
        self.assertEqual(self.object.accounts[2].name, "IncomeTaxExpense")
        self.assertEqual(self.object.accounts[2],
                         self.object.incometaxexpense_account)
        self.assertEqual(self.object.accounts[2].account_type,
                         AccountType.expense)
        self.assertEqual(self.object.accounts[3].name, "Sales")
        self.assertEqual(self.object.accounts[3],
                         self.object.sales_account)
        self.assertEqual(self.object.accounts[3].account_type,
                         AccountType.revenue)
        self.assertEqual(self.object.accounts[4].name, "CostOfSales")
        self.assertEqual(self.object.accounts[4],
                         self.object.costofsales_account)
        self.assertEqual(self.object.accounts[4].account_type,
                         AccountType.expense)
        self.assertEqual(self.object.accounts[5].name, "GrossProfit")
        self.assertEqual(self.object.accounts[5],
                         self.object.gross_profit_account)
        self.assertEqual(self.object.accounts[5].account_type,
                         AccountType.revenue)
        self.assertEqual(self.object.accounts[6].name, "IncomeSummary")
        self.assertEqual(self.object.accounts[6],
                         self.object.incomesummary_account)
        self.assertEqual(self.object.accounts[6].account_type,
                         AccountType.revenue)
        self.assertEqual(self.object.accounts[7].name, "RetainedEarnings")
        self.assertEqual(self.object.accounts[7],
                         self.object.retainedearnings_account)
        self.assertEqual(self.object.accounts[7].account_type,
                         AccountType.equity)
        self.assertEqual(self.object.tax_payment_account, "Bank")

    def test_create_account(self):
        orig_length = len(self.object.accounts)
        new_acc = self.object.create_account("TestA",
                                             description="TestA_Desc",
                                             number="011",
                                             account_type=AccountType.equity)
        self.assertEqual(new_acc.name, "TestA")
        self.assertEqual(new_acc.description, "TestA_Desc")
        self.assertEqual(new_acc.number, "011")
        self.assertEqual(new_acc.account_type, AccountType.equity)

        self.assertEqual(new_acc, self.object.accounts[orig_length])

    def test_remove_account(self):
        orig_length = len(self.object.accounts)
        self.object.create_account("TestA",
                                   description="TestA_Desc",
                                   number="011",
                                   account_type=AccountType.equity)
        self.object.remove_account("TestA")
        self.assertEqual(len(self.object.accounts), orig_length)

    def test_get_account(self):
        self.object.create_account("TestA",
                                   description="TestA_Desc",
                                   number="010",
                                   account_type=AccountType.equity)
        acc = self.object.create_account("TestB",
                                         description="TestB_Desc",
                                         number="020",
                                         account_type=AccountType.asset)
        sub_acc = acc.create_account("TestB1",
                                     description="TestB1_Desc",
                                     number="010")
        sub_acc.create_account("TestB1.1",
                               description="TestB1.1_Desc",
                               number="010")
        orig = sub_acc.create_account("TestB1.2",
                                      description="TestB1.1_Desc",
                                      number="011")
        result = self.object.get_account("TestB1.2")

        self.assertEqual(result.name, orig.name)
        self.assertEqual(result.description, orig.description)
        self.assertEqual(result.number, orig.number)

    def test_get_account_and_decendants(self):
        # Set up this test.
        self.sales_fish_acc = self.object.accounts[3].create_account(
            "SalesFish",
            description="Sales of Fish",
            number="010")
        self.sales_barracuda_acc = self.sales_fish_acc.create_account(
            "SalesBarracuda",
            description="Sales of Barracudas",
            number="010")
        self.sales_nemo_acc = self.sales_fish_acc.create_account(
            "SalesNemo",
            description="Sales of Nemos",
            number="020")
        # perform the test.
        result = []
        self.object.get_account_and_decendants(
            self.object.accounts[3], result)
        self.assertEqual(len(result), 4)
        self.assertEqual(result[0], self.object.accounts[3])
        self.assertEqual(result[1], self.sales_fish_acc)
        self.assertEqual(result[2], self.sales_barracuda_acc)
        self.assertEqual(result[3], self.sales_nemo_acc)

# =============================================================================
# Display documentation and run tests.
# =============================================================================
# os.system("cls")

# help(TransactionTemplate)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
This module provides a general ledger structure class which
structures a general ledger's accounts.\n

@name: general ledger structure
@author: Ex Mente Technologies (Pty) Ltd
"""

from auxi.core.namedobject import NamedObject
from auxi.modeling.financial.des.generalledgeraccount import AccountType
from auxi.modeling.financial.des.generalledgeraccount import GeneralLedgerAccount

__version__ = "0.2.0"

# TODO: initialise accounts by json file
# TODO: get_account by path.


class GeneralLedgerStructure(NamedObject):
    """Represents the account structure of a general ledger."""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, description=None):
        """Initialise the object."""
        super().__init__(name, description)
        self.accounts = []
        self.create_account(
            "Bank",
            description="Bank",
            number="010",
            account_type=AccountType.asset)
        self.incometaxpayable_account = self.create_account(
            "IncomeTaxPayable",
            description="IncomeTaxPayable",
            number="010",
            account_type=AccountType.liability)
        self.incometaxexpense_account = self.create_account(
            "IncomeTaxExpense",
            description="IncomeTaxExpense",
            number="010",
            account_type=AccountType.expense)
        self.sales_account = self.create_account(
            "Sales",
            description="Sales",
            number="010",
            account_type=AccountType.revenue)
        self.costofsales_account = self.create_account(
            "CostOfSales",
            description="CostOfSales",
            number="010",
            account_type=AccountType.expense)
        self.gross_profit_account = self.create_account(
            "GrossProfit",
            description="GrossProfit",
            number="010",
            account_type=AccountType.revenue)
        self.incomesummary_account = self.create_account(
            "IncomeSummary",
            description="IncomeSummary",
            number="010",
            account_type=AccountType.revenue)
        self.retainedearnings_account = self.create_account(
            "RetainedEarnings",
            description="RetainedEarnings",
            number="010",
            account_type=AccountType.equity)
        self.tax_payment_account = "Bank"

    def create_account(self, name, description=None, number=None,
                       account_type=AccountType.revenue):
        """Create an account in the general ledger structure.

        :param name: The account name.
        :param description: The account description.
        :param number: The account number.
        :param account_type: The account type.

        :returns: The created account.
        """
        new_account = GeneralLedgerAccount(name, description, number,
                                           account_type)
        self.accounts.append(new_account)
        return new_account

    def remove_account(self, name):
        """Remove an account from the general ledger's accounts.

        :param name: The name of the account to remove.
        """
        acc_to_remove = None
        for a in self.accounts:
            if a.name == name:
                acc_to_remove = a
        if acc_to_remove is not None:
            self.accounts.remove(acc_to_remove)

    def _get_account_from_child(self, accounts, account_name):
        for a in accounts:
            if a.name == account_name:
                return a
            else:
                result = self._get_account_from_child(a.accounts, account_name)
                if result is not None:
                    return result
        return None

    def get_account(self, account_name):
        """Retrieves an account from the general ledger structure
           given the account name.

        :param account_name: The account name.

        :returns: The requested account, if found, else None.
        """
        for a in self.accounts:
            if a.name == account_name:
                return a
            else:
                result = self._get_account_from_child(a.accounts, account_name)
                if result is not None:
                    return result
        return None

    def get_account_and_decendants(self, account, result):
        result.append(account)
        for child in account.accounts:
            self.get_account_and_decendants(child, result)



"""auxi.modeling.process.materials.thermochemistry Package

The purpose of this package is to provide classes and functions for the the auxi.modeling.process.materials.thermochemistry namespace"""


__version__ = "0.2.0"



# -*- coding: utf-8 -*-
"""
This module provides testing code for the auxi.modeling.financial.des.generalledgeraccount module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
from auxi.modeling.financial.des.generalledgeraccount import AccountType
from auxi.modeling.financial.des.generalledgeraccount import GeneralLedgerAccount

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class TestAllFunctions(unittest.TestCase):
    """
        Tester for the auxi.modeling.financial.des.generalledgeraccount class.
    """

    def setUp(self):
        self.object = GeneralLedgerAccount("NameA",
                                           description="DescriptionA",
                                           number="010",
                                           account_type=AccountType.asset)

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.number, "010")
        self.assertEqual(self.object.account_type, AccountType.asset)

    def test_create_account(self):
        new_account = self.object.create_account("TestA",
                                                 description="TestA_Desc",
                                                 number="011")
        self.assertEqual(new_account.name, "TestA")
        self.assertEqual(new_account.description, "TestA_Desc")
        self.assertEqual(new_account.number, "011")
        self.assertEqual(new_account.account_type, self.object.account_type)

        self.assertEqual(new_account, self.object.accounts[0])

    def test_remove_account(self):
        num_accounts = len(self.object.accounts)
        self.object.create_account("TestA",
                                   description="TestA_Desc",
                                   number="011")
        self.object.remove_account("TestA")
        self.assertEqual(len(self.object.accounts), num_accounts)

# =============================================================================
# Display documentation and run tests.
# =============================================================================
# os.system("cls")

# help(TransactionTemplate)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
This module provides a number of functions for doing thermochemical
calculations.

compounds: A dictionary of chemical compounds with thermochemical data loaded
from data files.

@author: Ex Mente Technologies (Pty) Ltd
"""

import os
import sys
import glob
from auxi.tools.chemistry import stoichiometry as stoich
from auxi.tools.chemistry.compound import Compound

__version__ = "0.2.0"


# =============================================================================
# Private functions.
# =============================================================================

def _get_default_data_path():
    """Calculate the default path in which thermochemical data is stored.

    :returns: Default path."""

    module_path = os.path.dirname(sys.modules[__name__].__file__)
    data_path = os.path.join(module_path, r"data/thermo")
    data_path = os.path.abspath(data_path)
    return data_path


def _read_compound_from_factsage_file(file_name):
    """Build a dictionary containing the thermochemical data of a compound by
    reading the data from a file.

    :param file_name: Name of file to read the data from.

    :returns: Dictionary containing compound data."""

    f = open(file_name, "r")
    lines = f.readlines()
    f.close()
    compound = dict()
    compound["Formula"] = lines[0].split(" ")[1]
    compound["Phases"] = dict()
    started = False
    phaseold = "zz"
    recordold = "0"
    for line in lines:
        if started:
            if line.startswith("_"):  # line indicating end of data
                break
            line = line.replace(" 298 ", " 298.15 ")
            line = line.replace(" - ", " ")
            while '  ' in line:
                line = line.replace("  ", " ")
            line = line.replace(" \n", "")
            line = line.replace("\n", "")
            strings = line.split(" ")
            if len(strings) < 2:  # empty line
                continue
            phase = strings[0]
            if phase != phaseold:  # new phase detected
                phaseold = phase
                compound["Phases"][phase] = dict()
                compound["Phases"][phase]["Symbol"] = phase
                compound["Phases"][phase]["DHref"] = float(strings[2])
                compound["Phases"][phase]["Sref"] = float(strings[3])
                compound["Phases"][phase]["Cp_records"] = dict()
                record = strings[1]
                if record != recordold:  # new record detected
                    recordold = record
                    Tmax = float(strings[len(strings) - 1])
                    compound["Phases"][phase]["Cp_records"][Tmax] = dict()
                    compound["Phases"][phase]["Cp_records"][Tmax]["Tmin"] = float(strings[len(strings) - 2])
                    compound["Phases"][phase]["Cp_records"][Tmax]["Tmax"] = float(strings[len(strings) - 1])
                    compound["Phases"][phase]["Cp_records"][Tmax]["Terms"] = list()
                    t = dict()
                    t["Coefficient"] = float(strings[4])
                    t["Exponent"] = float(strings[5])
                    compound["Phases"][phase]["Cp_records"][Tmax]["Terms"].append(t)
                    if len(strings) == 10:
                        t = dict()
                        t["Coefficient"] = float(strings[6])
                        t["Exponent"] = float(strings[7])
                        compound["Phases"][phase]["Cp_records"][Tmax]["Terms"].append(t)
                else:  # old record detected
                    t = dict()
                    t["Coefficient"] = float(strings[2])
                    t["Exponent"] = float(strings[3])
                    compound["Phases"][phase]["Cp_records"][Tmax]["Terms"].append(t)
                    if len(strings) == 8:
                        t = dict()
                        t["Coefficient"] = float(strings[4])
                        t["Exponent"] = float(strings[5])
                        compound["Phases"][phase]["Cp_records"][Tmax]["Terms"].append(t)
            else:  # old phase detected
                record = strings[1]
                if record != recordold:  # new record detected
                    recordold = record
                    Tmax = float(strings[len(strings) - 1])
                    compound["Phases"][phase]["Cp_records"][Tmax] = dict()
                    compound["Phases"][phase]["Cp_records"][Tmax]["Tmin"] = float(strings[len(strings) - 2])
                    compound["Phases"][phase]["Cp_records"][Tmax]["Tmax"] = float(strings[len(strings) - 1])
                    compound["Phases"][phase]["Cp_records"][Tmax]["Terms"] = list()
                    t = dict()
                    t["Coefficient"] = float(strings[2])
                    t["Exponent"] = float(strings[3])
                    compound["Phases"][phase]["Cp_records"][Tmax]["Terms"].append(t)
                    if len(strings) == 8:
                        t = dict()
                        t["Coefficient"] = float(strings[4])
                        t["Exponent"] = float(strings[5])
                        compound["Phases"][phase]["Cp_records"][Tmax]["Terms"].append(t)
                else:  # old record detected
                    t = dict()
                    t["Coefficient"] = float(strings[2])
                    t["Exponent"] = float(strings[3])
                    compound["Phases"][phase]["Cp_records"][Tmax]["Terms"].append(t)
                    if len(strings) == 8:
                        t = dict()
                        t["Coefficient"] = float(strings[4])
                        t["Exponent"] = float(strings[5])
                        compound["Phases"][phase]["Cp_records"][Tmax]["Terms"].append(t)
        if line.startswith("_"):  # line indicating the start of the data
            started = True
    return compound


def _split_compound_string(compound_string):
    """Split a compound's combined formula and phase into separate strings for
    the formula and phase.

    :param compound_string: Formula and phase of a chemical compound, e.g. \
    "SiO2[S1].

    :returns: Formula of chemical compound.
    :returns: Phase of chemical compound."""

    formula = compound_string.replace("]", "").split("[")[0]
    phase = compound_string.replace("]", "").split("[")[1]
    return formula, phase


def _finalise_result(compound, value, mass):
    """Convert the value to its final form by unit conversions and multiplying \
    by mass.

    :param compound: Compound object.
    :param value:    Value to be finalised. [J/mol]
    :param mass:     Mass of compound. [kg]

    :returns: Finalised value. [kWh]"""

    result = value / 3.6E6  # J/x -> kWh/x
    result = result / compound.molar_mass  # x/mol -> x/kg
    result = result * mass  # x/kg -> x
    return result


# =============================================================================
# Public functions.
# =============================================================================

def load_data(path=""):
    """Load all the thermochemical data files located at a path.

    :param path: Path at which the data files are located."""

    compounds.clear()

    if path == "":
        path = default_data_path
    if not os.path.exists(path):
        raise Exception("The path does not exist.")

    files = glob.glob(os.path.join(path, "Compound_*.txt"))

    for file in files:
        compound = Compound(_read_compound_from_factsage_file(file))
        compounds[compound.formula] = compound


def list_compounds():
    """List all compounds that are currently loaded in the thermo module, and
    their phases."""

    print("Compounds currently loaded in the thermo module:")
    for compound in sorted(compounds.keys()):
        phases = compounds[compound].get_phase_list()
        for phase in phases:
            print(compound + "[" + phase + "]")


def molar_mass(compound):
    """Determine the molar mass of a chemical compound.

    :param compound: Formula of a chemical compound, e.g. "Fe2O3".

    :returns: Molar mass. [kg/mol]"""

    return stoich.molecular_mass(compound) / 1000.0


def Cp(compound_string, temperature, mass=1.0):
    """Calculate the heat capacity of the compound for the specified
    temperature and mass.

    :param compound_string: Formula and phase of chemical compound, e.g. \
    "Fe2O3[S1]".
    :param temperature:     [°C]
    :param mass:            [kg]

    :returns: Heat capacity. [kWh/K]"""

    formula, phase = _split_compound_string(compound_string)
    temperature_K = temperature + 273.15
    compound = compounds[formula]
    result = compound.Cp(phase, temperature_K)
    return _finalise_result(compound, result, mass)


def H(compound_string, temperature, mass=1.0):
    """Calculate the enthalpy of the compound for the specified temperature
    and mass.

    :param compound_string: Formula and phase of chemical compound, e.g. \
    "Fe2O3[S1]".
    :param temperature:     [°C]
    :param mass:            [kg]

    :returns: Enthalpy. [kWh]"""

    formula, phase = _split_compound_string(compound_string)
    temperature_K = temperature + 273.15
    compound = compounds[formula]
    result = compound.H(phase, temperature_K)
    return _finalise_result(compound, result, mass)


def S(compound_string, temperature, mass=1.0):
    """Calculate the entropy of the compound for the specified temperature and
    mass.

    :param compound_string: Formula and phase of chemical compound, e.g. \
    "Fe2O3[S1]".
    :param temperature:     [°C]
    :param mass:            [kg]

    :returns: Entropy. [kWh/K]"""

    formula, phase = _split_compound_string(compound_string)
    temperature_K = temperature + 273.15
    compound = compounds[formula]
    result = compound.S(phase, temperature_K)
    return _finalise_result(compound, result, mass)


def G(compound_string, temperature, mass=1.0):
    """Calculate the Gibbs free energy of the compound for the specified
    temperature and mass.

    :param compound_string: Formula and phase of chemical compound, e.g. \
    "Fe2O3[S1]".
    :param temperature:     [°C]
    :param mass:            [kg]


    :returns: Gibbs free energy. [kWh]"""

    formula, phase = _split_compound_string(compound_string)
    temperature_K = temperature + 273.15
    compound = compounds[formula]
    result = compound.G(phase, temperature_K)
    return _finalise_result(compound, result, mass)


# =============================================================================
# Public variables.
# =============================================================================

compounds = {}
default_data_path = _get_default_data_path()
load_data()



# -*- coding: utf-8 -*-
"""
This module provides testing code for the chemistry material module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
import os
import numpy
from auxi.modeling.process.materials.chemistry import material
from auxi.modeling.process.materials.chemistry.material import Material
from auxi.modeling.process.materials.chemistry.material import MaterialPackage

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class TestMaterial(unittest.TestCase):
    """Tester for the auxi.modeling.process.materials.chemistry.material.Material class."""

    def setUp(self):
        self.material = Material("material",
                                 os.path.join(material.DEFAULT_DATA_PATH,
                                 r"chemmaterial.test.ilmenite.txt"))

    def test_constructor(self):
        self.assertEqual(self.material.name, "material")
        self.assertEqual(len(self.material.compounds), 14)
        self.assertEqual(self.material.compound_count, 14)
        self.assertEqual(len(self.material.assays), 3)

    def test_get_compound_index(self):
        self.assertEqual(self.material.get_compound_index("Al2O3"), 0)
        self.assertEqual(self.material.get_compound_index("K2O"), 6)
        self.assertEqual(self.material.get_compound_index("P4O10"), 10)
        self.assertEqual(self.material.get_compound_index("V2O5"), 13)

    def test_create_empty_assay(self):
        empty_assay = self.material.create_empty_assay()
        self.assertEqual(len(empty_assay), 14)
        self.assertEqual(empty_assay.sum(), 0.0)
        self.assertEqual(empty_assay.sum(), 0.0)
        self.assertEqual(empty_assay.sum(), 0.0)

    def test_add_assay(self):
        new_assay = self.material.create_empty_assay()
        new_assay[0] = 0.5
        new_assay[2] = 0.5
        self.material.add_assay("new_assay", new_assay)
        self.assertEqual(
            numpy.all(self.material.assays["new_assay"] == new_assay),
            True)

    def test_get_assay_total(self):
        self.assertAlmostEqual(self.material.get_assay_total("IlmeniteA"),
                               0.99791999999999992,
                               places=10)
        self.assertAlmostEqual(self.material.get_assay_total("IlmeniteB"),
                               0.99760999999999989,
                               places=10)
        self.assertAlmostEqual(self.material.get_assay_total("IlmeniteC"),
                               1.0000200000000001,
                               places=10)

    def test_create_package(self):
        package = self.material.create_package("IlmeniteA", 123.456, True)
        self.assertEqual(package.get_mass(), 123.456)


class TestMaterialPackage(unittest.TestCase):
    """Tester for the auxi.modeling.process.materials.chemistry.material.Material class."""
    def setUp(self):
        self.ilmenite = Material("ilmenite",
                                 os.path.join(material.DEFAULT_DATA_PATH,
                                 r"chemmaterial.test.ilmenite.txt"))
        self.ilmenite_package_a = self.ilmenite.create_package("IlmeniteA",
                                                             1234.5, True)
        self.ilmenite_package_b = self.ilmenite.create_package("IlmeniteB",
                                                             2345.6, True)
        self.ilmenite_package_c = self.ilmenite.create_package("IlmeniteC",
                                                             3456.7, True)
        self.reductant = Material("reductant",
                                  os.path.join(material.DEFAULT_DATA_PATH,
                                  r"chemmaterial.test.reductant.txt"))
        self.reductant_package_a = self.reductant.create_package("ReductantA",
                                                               123.45, True)
        self.mix = Material("mix", os.path.join(material.DEFAULT_DATA_PATH,
                                                r"chemmaterial.test.mix.txt"))

    def test_constructor(self):
        compound_masses = self.ilmenite.assays["IlmeniteB"] * 123.4 / \
                          self.ilmenite.assays["IlmeniteB"].sum()
        package = MaterialPackage(self.ilmenite, compound_masses)
        self.assertAlmostEqual(package.get_mass(), 123.4,
                               places=10)

    def test_add_operator_1(self):
        """other = MaterialPackage
        Test whether the add operator calculates the resulting package
        correctly. Results were checked against FactSage results. They are not
        exactly the same, since the magnetic and other non-cp contributions are
        omitted by the thermo module.
        """

        package_a_plus_b = self.ilmenite_package_a + self.ilmenite_package_b
        package_a_plus_c = self.ilmenite_package_a + self.ilmenite_package_c
        package_b_plus_c = self.ilmenite_package_b + self.ilmenite_package_c
        package_a_plus_b_plus_c = self.ilmenite_package_a + \
                             self.ilmenite_package_b + \
                             self.ilmenite_package_c

        self.assertAlmostEqual(self.ilmenite_package_a.get_mass(), 1234.5,
                               places=10)

        self.assertAlmostEqual(self.ilmenite_package_b.get_mass(), 2345.6,
                               places=10)

        self.assertAlmostEqual(self.ilmenite_package_c.get_mass(), 3456.7,
                               places=10)

        self.assertAlmostEqual(package_a_plus_b.get_mass(), 3580.0999999999999,
                               places=10)

        self.assertAlmostEqual(package_a_plus_c.get_mass(), 4691.1999999999989,
                               places=10)

        self.assertAlmostEqual(package_b_plus_c.get_mass(), 5802.2999999999993,
                               places=10)

        self.assertAlmostEqual(package_a_plus_b_plus_c.get_mass(), 7036.8,
                               places=10)

    def test_add_operator_2(self):
        mix_package = self.mix.create_package(None, 0.0)
        mix_package = mix_package + self.ilmenite_package_a
        mix_package = mix_package + self.reductant_package_a

        self.assertAlmostEqual(mix_package.get_mass(),
                               self.ilmenite_package_a.get_mass() +
                               self.reductant_package_a.get_mass(),
                               places=10)

        self.assertRaises(Exception, self.add_incompatible_packages)

    def add_incompatible_packages(self):
        result = self.ilmenite_package_a + self.reductant_package_a
        result = result * 1.0

    def test_add_operator_3(self):
        """other = tuple (compound, mass)
        Test whether the add operator calculates the resulting package
        correctly. Results were checked against FactSage results. They are not
        exactly the same, since the magnetic and other non-cp contributions are
        omitted by the thermo module.
        """

        packageAplusAl2O3 = self.ilmenite_package_a + ("Al2O3", 123.4)

        self.assertEqual(packageAplusAl2O3.get_mass(), 1357.9)

    def test_extract_1(self):
        temp_package_a = self.ilmenite_package_a.clone()
        mass = 432.1
        diff_package = temp_package_a.extract(mass)

        self.assertAlmostEqual(temp_package_a.get_mass(),
                               self.ilmenite_package_a.get_mass() - mass,
                               places=10)

        self.assertAlmostEqual(diff_package.get_mass(), mass,
                               places=10)

    def test_subtract_operator_2(self):
        temp_package_a = self.ilmenite_package_a.clone()
        compound = "TiO2"
        mass = 123.4
        diff_package = temp_package_a.extract((compound, mass))

        self.assertAlmostEqual(temp_package_a.get_mass(),
                               self.ilmenite_package_a.get_mass() - mass,
                               places=10)
        self.assertAlmostEqual(temp_package_a.get_compound_mass(compound),
                               self.ilmenite_package_a.get_compound_mass(compound) -
                               mass,
                               places=10)

        self.assertEqual(diff_package.get_mass(), mass)
        self.assertEqual(diff_package.get_compound_mass(compound), mass)

    def test_subtract_operator_3(self):
        temp_package_a = self.ilmenite_package_a.clone()
        compound = "TiO2"
        mass = temp_package_a.get_compound_mass(compound)
        diff_package = temp_package_a.extract(compound)

        self.assertAlmostEqual(temp_package_a.get_mass(),
                               self.ilmenite_package_a.get_mass() - mass,
                               places=10)
        self.assertAlmostEqual(temp_package_a.get_compound_mass(compound),
                               self.ilmenite_package_a.get_compound_mass(compound) -
                               mass,
                               places=10)

        self.assertEqual(diff_package.get_mass(), mass)
        self.assertEqual(diff_package.get_compound_mass(compound), mass)

    def test_multiply_operator(self):
        temp_package_a = self.ilmenite_package_a.clone()

        mul_package_1 = temp_package_a * 0.0
        self.assertEqual(mul_package_1.get_mass(), 0.0)
        self.assertTrue(numpy.all(mul_package_1.compound_masses ==
                                  temp_package_a.compound_masses * 0.0))

        mul_package_2 = temp_package_a * 1.0
        self.assertAlmostEqual(mul_package_2.get_mass(),
                               temp_package_a.get_mass(),
                               places=10)
        self.assertTrue(numpy.all(mul_package_2.compound_masses ==
                                  temp_package_a.compound_masses))

        mul_package_2 = temp_package_a * 123.4
        self.assertAlmostEqual(mul_package_2.get_mass(),
                               temp_package_a.get_mass() * 123.4,
                               places=10)
        self.assertTrue(numpy.all(mul_package_2.compound_masses ==
                                  temp_package_a.compound_masses * 123.4))

    def test_clone(self):
        clone = self.ilmenite_package_a.clone()

        self.assertEqual(clone.get_mass(), self.ilmenite_package_a.get_mass())
        self.assertTrue(numpy.all(clone.compound_masses ==
                                  self.ilmenite_package_a.compound_masses))

    def test_get_mass(self):
        self.assertAlmostEqual(self.ilmenite_package_a.get_mass(), 1234.5,
                               places=10)
        self.assertAlmostEqual(self.ilmenite_package_b.get_mass(), 2345.6,
                               places=10)
        self.assertAlmostEqual(self.ilmenite_package_c.get_mass(), 3456.7,
                               places=10)

    def test_get_assay(self):
        self.assertTrue(numpy.all(self.ilmenite_package_a.get_assay() -
                                  self.ilmenite.assays["IlmeniteA"] /
                                  self.ilmenite.assays["IlmeniteA"].sum() <
                                  1.0E-16))
        self.assertTrue(numpy.all(self.ilmenite_package_a.get_assay() -
                                  self.ilmenite.assays["IlmeniteA"] /
                                  self.ilmenite.assays["IlmeniteA"].sum() >
                                  -1.0E-16))

    def test_get_compound_mass(self):
        assay = "IlmeniteA"
        compound = "TiO2"
        for compound in self.ilmenite.compounds:
            index = self.ilmenite.get_compound_index(compound)
            mass = 1234.5 * self.ilmenite.assays[assay][index] / \
                   self.ilmenite.get_assay_total(assay)
            self.assertEqual(
                self.ilmenite_package_a.get_compound_mass(compound),
                mass)

    def test_get_element_masses(self):
        x = self.ilmenite_package_a.get_element_masses()
        y = self.ilmenite_package_a.get_element_mass("Ti")


# =============================================================================
# Display documentation and run tests.
# =============================================================================
#os.system("cls")

#help(Material)
#help(MaterialPackage)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
This module provides testing code for the
auxi.modeling.business.activity module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
from datetime import datetime
from auxi.modeling.business.clock import Clock
from auxi.modeling.business.activity import Activity

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class TestAllFunctions(unittest.TestCase):
    """
      Tester for the auxi.modeling.business.activity class.
    """

    def setUp(self):
        self.object = Activity("NameA",
                               description="DescriptionA",
                               start=datetime(2016, 2, 1),
                               end=datetime(2017, 2, 1),
                               interval=3)

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.start_datetime, datetime(2016, 2, 1))
        self.assertEqual(self.object.end_datetime, datetime(2017, 2, 1))
        self.assertEqual(self.object.interval, 3)

    def test_get_path(self):
        self.object.set_parent_path("entityA/componentA")
        self.assertEqual(self.object.path, "entityA/componentA/NameA")

    def test_set_parent_path(self):
        self.object.set_parent_path("entityA/componentA")
        self.assertEqual(self.object.path, "entityA/componentA/NameA")

    def test_set_name(self):
        self.object.set_parent_path("entityA/componentA")
        self.object.name = "NameAt"
        self.assertEqual(self.object.name, "NameAt")
        self.assertEqual(self.object.path, "entityA/componentA/NameAt")

    def test__meet_exection_criteria(self):
        clock = Clock("NameA", start_datetime=datetime(2016, 1, 1))
        self.object.prepare_to_run(clock, 13)
        self.assertEqual(self.object._meet_execution_criteria(3), False)
        self.assertEqual(self.object._meet_execution_criteria(5), True)
        self.assertEqual(self.object._meet_execution_criteria(40), False)

    def test_prepare_to_run(self):
        clock = Clock("NameA", start_datetime=datetime(2016, 1, 1))
        self.object.prepare_to_run(clock, 18)
        self.assertEqual(self.object.start_period_ix, 2)
        self.assertEqual(self.object.end_period_ix, 14)


# =============================================================================
# Display documentation and run tests.
# =============================================================================
# os.system("cls")

# help(Activity)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Created on Tue May  6 08:09:58 2014

@author: Ex Mente (Pty) Ltd
"""
__version__ = "0.2.0"

# TODO: Fix the comment outed tests

import os
import unittest

from auxi.core.object_test import TestAllFunctions as object_test_all
from auxi.core.namedobject_test import TestAllFunctions as namedobject_test_all

#from auxi.tools.chemistry.stoichiometry_test import TestAllFunctions as stoich_test_all
from auxi.tools.chemistry.thermochemistry_test import TestAllFunctions as thermo_test_all

from auxi.modeling.financial.des.transactiontemplate_test import TestAllFunctions as transactiontemplate_test_all
from auxi.modeling.financial.des.transaction_test import TestAllFunctions as transaction_test_all
from auxi.modeling.financial.des.generalledgeraccount_test import TestAllFunctions as generalledgeraccount_test_all
from auxi.modeling.financial.des.generalledgerstructure_test import TestAllFunctions as generalledgerstructure_test_all
from auxi.modeling.financial.des.generalledger_test import TestAllFunctions as generalledger_test_all
from auxi.modeling.financial.des.currency_test import TestAllFunctions as currency_test_all
from auxi.modeling.financial.des.currencytable_test import TestAllFunctions as currencytable_test_all

from auxi.modeling.financial.tax.rule_test import TestAllFunctions as taxrule_test_all
from auxi.modeling.financial.tax.ruleset_test import TestAllFunctions as taxruleset_test_all
from auxi.modeling.financial.tax.salesrule_test import TestAllFunctions as salestaxrule_test_all
from auxi.modeling.financial.tax.incomerule_test import TestAllFunctions as incometaxrule_test_all
from auxi.modeling.financial.tax.capitalgainsrule_test import TestAllFunctions as capitalgainstaxrule_test_all

from auxi.modeling.business.clock_test import TestAllFunctions as clock_test_all
from auxi.modeling.business.activity_test import TestAllFunctions as activity_test_all
from auxi.modeling.business.basicactivity_test import TestAllFunctions as basicactivity_test_all
from auxi.modeling.business.component_test import TestAllFunctions as component_test_all
from auxi.modeling.business.entity_test import TestAllFunctions as entity_test_all
from auxi.modeling.business.timebasedmodel_test import TestAllFunctions as timebasedmodel_test_all

from auxi.modeling.process.materials.chemistry.material_test import TestMaterial as chem_test_material
from auxi.modeling.process.materials.chemistry.material_test import TestMaterialPackage as chem_test_material_package
#from auxi.modeling.process.materials.thermochemistry.material_test import TestMaterial as thermo_test_material
#from auxi.modeling.process.materials.thermochemistry.material_test import TestMaterialPackage as thermo_test_material_package
from auxi.modeling.process.materials.psd.material_test import TestMaterial as psd_test_material
from auxi.modeling.process.materials.psd.material_test import TestMaterialPackage as psd_test_material_package


if __name__ == '__main__':
    unittest.main()

#def run_all():
#    # os.system("cls")
#    print(__name__)
#    unittest.main(module=__name__)



# -*- coding: utf-8 -*-
"""
This module provides testing code for the
auxi.modeling.business.basicactivity module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
from datetime import datetime
from auxi.modeling.business.clock import Clock
from auxi.modeling.business.basicactivity import BasicActivity
from auxi.modeling.financial.des.generalledger import GeneralLedger
from auxi.modeling.financial.des.generalledgerstructure import GeneralLedgerStructure
from auxi.modeling.financial.des.transactiontemplate import TransactionTemplate

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class TestAllFunctions(unittest.TestCase):
    """
      Tester for the auxi.modeling.business.basicactivity class.
    """

    def setUp(self):
        self.tx_template = TransactionTemplate("NameA",
                                               description="DescriptionA",
                                               dt_account="Bank",
                                               cr_account="Sales")
        self.object = BasicActivity("NameA",
                                    description="DescriptionA",
                                    start=datetime(2016, 2, 1),
                                    end=datetime(2017, 2, 1),
                                    interval=3,
                                    amount=5000,
                                    tx_template=self.tx_template)

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.start_datetime, datetime(2016, 2, 1))
        self.assertEqual(self.object.end_datetime, datetime(2017, 2, 1))
        self.assertEqual(self.object.interval, 3)
        self.assertEqual(self.object.amount, 5000)
        self.assertEqual(self.object.tx_template, self.tx_template)

    def test__meet_exection_criteria(self):
        clock = Clock("NameA", start_datetime=datetime(2016, 1, 1))
        self.object.prepare_to_run(clock, 13)
        self.assertEqual(self.object._meet_execution_criteria(5), True)
        self.object.amount = 0
        self.assertEqual(self.object._meet_execution_criteria(5), False)
        self.object.amount = 5000

    def test_run(self):
        structure = GeneralLedgerStructure("NameA", description="DescriptionA")
        gl = GeneralLedger("NameA", structure, description="DescriptionA")
        clock = Clock("NameA", start_datetime=datetime(2016, 1, 1))
        clock.tick()
        clock.tick()
        self.object.prepare_to_run(clock, 20)
        self.object.run(clock, gl)
        self.assertEqual(len(gl.transactions), 1)
        self.assertEqual(gl.transactions[0].amount, 5000)


# =============================================================================
# Display documentation and run tests.
# =============================================================================
# os.system("cls")

# help(BasicActivity)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
This module provides testing code for the stoichiometry module.

@author: Ex Mente Technologies (Pty) Ltd
"""


import os
from auxi.tools.chemistry import stoichiometry as stoich
# import stoich
import unittest

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class TestAllFunctions(unittest.TestCase):

    def test_amount(self):
        self.assertEqual(stoich.amount("FeO", 1.0), 0.01391896932815919)

    def test_mass(self):
        self.assertEqual(stoich.mass("FeO", 1.0), 71.8444)

    def test_convert_compound(self):
        mm_Fe = stoich.molar_mass("Fe")
        mm_Fe2O3 = stoich.molar_mass("Fe2O3")

        mass_Fe = 1000.0
        mass_Fe2O3 = stoich.convert_compound(mass_Fe, "Fe", "Fe2O3", "Fe")
        self.assertEqual(mass_Fe2O3, mass_Fe / mm_Fe / 2.0 * mm_Fe2O3)
        mass_Fe2O3 = stoich.convert_compound(mass_Fe, "Fe", "Fe2O3", "O")
        self.assertEqual(mass_Fe2O3, 0.0)

        mm_FeTiO3 = stoich.molar_mass("FeTiO3")
        mm_FeO = stoich.molar_mass("FeO")
        mm_TiO2 = stoich.molar_mass("TiO2")

        mass_FeTiO3 = 1000.0
        mass_FeO = stoich.convert_compound(mass_FeTiO3, "FeTiO3", "FeO", "Fe")
        self.assertEqual(mass_FeO, mass_FeTiO3 / mm_FeTiO3 * mm_FeO)
        mass_TiO2 = stoich.convert_compound(mass_FeTiO3, "FeTiO3", "TiO2", "Ti")
        self.assertEqual(mass_TiO2, mass_FeTiO3 / mm_FeTiO3 * mm_TiO2)
        self.assertEqual(mass_FeO + mass_TiO2, mass_FeTiO3)

    def test_element_mass_fraction(self):
        self.assertEqual(stoich.element_mass_fraction("FeO", "Fe"), 0.7773048421310499)
        self.assertEqual(stoich.element_mass_fraction("FeO", "O"), 0.22269515786895014)
        self.assertEqual(stoich.element_mass_fraction("FeO", "Si"), 0.0)
        self.assertEqual(stoich.element_mass_fraction("Fe2O3", "Fe"), 0.699425505453753)
        self.assertEqual(stoich.element_mass_fraction("Fe2O3", "O"), 0.300574494546247)
        self.assertEqual(stoich.element_mass_fraction("Fe2O3", "Mn"), 0.0)
        self.assertEqual(stoich.element_mass_fraction("SiO2", "Si"), 0.4674349206032192)
        self.assertEqual(stoich.element_mass_fraction("SiO2", "O"), 0.5325650793967809)
        self.assertEqual(stoich.element_mass_fraction("SiO2", "Fe"), 0.0)
        self.assertEqual(stoich.element_mass_fraction("Ca(OH)2", "Ca"), 0.5409171324346751)
        self.assertEqual(stoich.element_mass_fraction("Ca(OH)2", "O"), 0.43187532155673136)
        self.assertEqual(stoich.element_mass_fraction("Ca(OH)2", "H"), 0.027207546008593562)
        self.assertEqual(stoich.element_mass_fraction("Ca(OH)2", "Fe"), 0.0)

    def test_element_mass_fractions(self):
        self.assertEqual(list(stoich.element_mass_fractions("FeO", ["Fe", "O", "Si"])), [0.7773048421310499, 0.22269515786895014, 0.0])
        self.assertEqual(list(stoich.element_mass_fractions("Fe2O3", ["Fe", "O", "Mn"])), [0.699425505453753, 0.300574494546247, 0.0])
        self.assertEqual(list(stoich.element_mass_fractions("SiO2", ["Si", "O", "Fe"])), [0.4674349206032192, 0.5325650793967809, 0.0])
        self.assertEqual(list(stoich.element_mass_fractions("Ca(OH)2", ["Ca", "O", "H", "Fe"])), [0.5409171324346751, 0.43187532155673136, 0.027207546008593562, 0.0])

    def test_elements(self):
        compounds = ["Ar", "Fe2O3", "SiO2", "Al2O3", "SO3", "CaO", "Fe", "Mn2O3"]
        elements = stoich.elements(compounds)
        self.assertEqual(elements, {'Al', 'Ar', 'Ca', 'Fe', 'Mn', 'O', 'S', 'Si'})
        compounds = ["CO2"]
        elements = stoich.elements(compounds)
        self.assertEqual(elements, {'O', 'C'})

    def test_molar_mass(self):
        self.assertEqual(stoich.molar_mass("FeO"), 71.8444)
        self.assertEqual(stoich.molar_mass("Fe2O3"), 159.6882)
        self.assertEqual(stoich.molar_mass("SiO2"), 60.0843)
        self.assertEqual(stoich.molar_mass("Ca(OH)2"), 74.09268)

    def test_stoichiometry_coefficient(self):
        self.assertEqual(stoich.stoichiometry_coefficient("FeO", "Fe"), 1.0)
        self.assertEqual(stoich.stoichiometry_coefficient("FeO", "O"), 1.0)
        self.assertEqual(stoich.stoichiometry_coefficient("FeO", "Si"), 0.0)
        self.assertEqual(stoich.stoichiometry_coefficient("Fe2O3", "Fe"), 2.0)
        self.assertEqual(stoich.stoichiometry_coefficient("Fe2O3", "O"), 3.0)
        self.assertEqual(stoich.stoichiometry_coefficient("Fe2O3", "Mn"), 0.0)
        self.assertEqual(stoich.stoichiometry_coefficient("SiO2", "Si"), 1.0)
        self.assertEqual(stoich.stoichiometry_coefficient("SiO2", "O"), 2.0)
        self.assertEqual(stoich.stoichiometry_coefficient("SiO2", "Fe"), 0.0)
        self.assertEqual(stoich.stoichiometry_coefficient("Ca(OH)2", "Ca"), 1.0)
        self.assertEqual(stoich.stoichiometry_coefficient("Ca(OH)2", "O"), 2.0)
        self.assertEqual(stoich.stoichiometry_coefficient("Ca(OH)2", "H"), 2.0)
        self.assertEqual(stoich.stoichiometry_coefficient("Ca(OH)2", "Fe"), 0.0)

    def test_stoichiometry_coefficients(self):
        self.assertEqual(stoich.stoichiometry_coefficients("FeO", ["Fe", "O", "Si"]), [1.0, 1.0, 0.0])
        self.assertEqual(stoich.stoichiometry_coefficients("Fe2O3", ["Fe", "O", "Mn"]), [2.0, 3.0, 0.0])
        self.assertEqual(stoich.stoichiometry_coefficients("FeOFe2O3", ["Fe", "O", "Mn"]), [3.0, 4.0, 0.0])
        self.assertEqual(stoich.stoichiometry_coefficients("(FeO)(Fe2O3)", ["Fe", "O", "Mn"]), [3.0, 4.0, 0.0])
        self.assertEqual(stoich.stoichiometry_coefficients("SiO2", ["Si", "O", "Fe"]), [1.0, 2.0, 0.0])
        self.assertEqual(stoich.stoichiometry_coefficients("Ca(OH)2", ["Ca", "O", "H", "Fe"]), [1.0, 2.0, 2.0, 0.0])
        self.assertEqual(stoich.stoichiometry_coefficients("CaAl2(Si2O7)(OH)2·H2O", ["Al", "Ca", "Si", "O", "H"]), [2.0, 1.0, 2.0, 10.0, 4.0])


# =============================================================================
# Display documentation and run tests.
# =============================================================================

if __name__ == '__main__':
    os.system("cls")
    # help(stoich)
    unittest.main()



# -*- coding: utf-8 -*-
"""
This module provides a component class that represents a component
of an entity. A component has business activities.\n

@name: component
@author: Ex Mente Technologies (Pty) Ltd
"""

from auxi.core.namedobject import NamedObject
from auxi.modeling.financial.des.transactiontemplate import TransactionTemplate

__version__ = "0.2.0"

# TODO: initialise accounts by json file
# TODO: get_account by path.


class Component(NamedObject):
    """Represents an component class."""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, description=None):
        """Initialise the object.

        :param name: The name.
        :param description: The description.
        """
        self._parent_path = ""
        self.path = name
        self.components = []
        self.activities = []
        super().__init__(name, description=description)

    def set_parent_path(self, value):
        """Set the parent path and the path from the new parent path.

        :param value: The path to the object's parent
        """
        self._parent_path = value
        self.path = value + r'/' + self.name
        self._update_childrens_parent_path()

    def _update_childrens_parent_path(self):
        for c in self.components:
            c.set_parent_path(self.path)
        for a in self.activities:
            a.set_parent_path(self.path)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self.path = self._parent_path + r'/' + self.name
        self._update_childrens_parent_path()

    def create_component(self, name, description=None):
        """Create a sub component in the business component.

        :param name: The new component's name.
        :param description: The enw component's description.

        :returns: The created component.
        """
        new_comp = Component(name, description=description)
        self.components.append(new_comp)
        return new_comp

    def remove_component(self, name):
        """Remove a sub component from the component.

        :param name: The name of the component to remove.
        """
        component_to_remove = None
        for c in self.components:
            if c.name == name:
                component_to_remove = c
        if component_to_remove is not None:
            self.components.remove(component_to_remove)

    def prepare_to_run(self, clock, period_count):
        """Prepare the component for execution.

        :param clock: The clock containing the execution start time and execution period information.
        :param period_count: The total amount of periods this activity will be requested to be run for.
        """
        for c in self.components:
            c.prepare_to_run(clock, period_count)
        for a in self.activities:
            a.prepare_to_run(clock, period_count)

    def run(self, clock, generalLedger):
        """Execute the component at the current clock cycle.

        :param clock: The clock containing the current execution time and period information.
        :param generalLedger: The general ledger into which to create the transactions.
        """
        for c in self.components:
            c.run(clock, generalLedger)
        for a in self.activities:
            a.run(clock, generalLedger)



# -*- coding: utf-8 -*-
"""
This module contains a single class representing a Cp record of a chemical
compound phase.

@author: Ex Mente Technologies (Pty) Ltd
"""

import math
from auxi.core.object import Object

__version__ = "0.2.0"


class CpRecord(Object):
    """A Cp equation record for a compound phase over a specific temperature
    range.

    :param dictionary: Dictionary containing the data required to initialise \
    the phase."""

    def __init__(self, dictionary):
        self.Tmin = dictionary["Tmin"]

        """The minimum temperature of the range covered by this record. [K]"""
        self.Tmax = dictionary["Tmax"]
        """The maximum temperature of the range covered by this record. [K]"""

        self._coefficients = list()
        """The coefficiencts of the terms in the Cp equation."""

        self._exponents = list()
        """The exponents of the terms in the Cp equation."""

        for t in dictionary["Terms"]:
            self._coefficients.append(t["Coefficient"])
            self._exponents.append(t["Exponent"])


    def __str__(self):
        result = "\t\tCp RECORD:" + "\n"
        result = result + "\t\t\tTmin: " + str(self.Tmin) + "\n"
        result = result + "\t\t\tTmax: " + str(self.Tmax) + "\n"
        for i in range(len(self._coefficients)):
            result = result + "\t\t\t{:.8e}".format(self._coefficients[i]) + " "
            result = result + str(self._exponents[i]) + "\n"
        return result


    def Cp(self, temperature):
        """Calculate the heat capacity of the compound phase.

        :param temperature: [K]

        :returns: Heat capacity. [J/mol/K]"""

        result = 0.0
        for c, e in zip(self._coefficients, self._exponents):
            result = result + c * temperature ** e
        return result


    def H(self, temperature):
        """Calculate the portion of enthalpy of the compound phase covered by
        this Cp record.

        :param temperature: [K]

        :returns: Enthalpy. [J/mol]"""

        result = 0.0
        if temperature < self.Tmax:
            T = temperature
        else:
            T = self.Tmax
        Tref = self.Tmin
        for c, e in zip(self._coefficients, self._exponents):
            # Analytically integrate Cp(T).
            if e == -1.0:
                result = result + c * (math.log(T) - math.log(Tref))
            else:
                result = result + c * (T ** (e + 1.0) - Tref ** (e + 1.0)) / (e + 1.0)
        return result


    def S(self, temperature):
        """Calculate the portion of entropy of the compound phase covered by
        this Cp record.

        :param temperature: [K]

        :returns: Entropy. [J/mol/K]"""

        result = 0.0
        if temperature < self.Tmax:
            T = temperature
        else:
            T = self.Tmax
        Tref = self.Tmin
        for c, e in zip(self._coefficients, self._exponents):
            # Create a modified exponent to analytically integrate Cp(T)/T instead of Cp(T).
            e_modified = e - 1.0
            if e_modified == -1.0:
                result = result + c * (math.log(T) - math.log(Tref))
            else:
                result = result + c * (T ** (e_modified + 1.0) - Tref ** (e_modified + 1.0)) / (e_modified + 1.0)
        return result



# -*- coding: utf-8 -*-
"""
This module provides a class that converts raw assays for use in calculations
by normalising.

@author: Ex Mente Technologies (Pty) Ltd
"""
__version__ = "0.2.0"

from assayconversionaction import AssayConversionAction

class Normalise(AssayConversionAction):
    def run(self, material):
        pass


conversion = [Normalise(), ]
mat = Material()



# -*- coding: utf-8 -*-
"""
This module provides a material class that can do thermochemical calculations.

@author: Ex Mente Technologies (Pty) Ltd
"""

import os
import sys
import numpy
import copy
from auxi.core.object import Object
from auxi.core.namedobject import NamedObject
from auxi.tools.chemistry import stoichiometry as stoich
from auxi.tools.chemistry import thermochemistry as thermo

__version__ = "0.2.0"


class Material(NamedObject):
    """Represents a material consisting of multiple chemical compounds, having
    the ability to do thermochemical calculations.

    :param name: A name for the material.
    :param file_path: The location of the file containing the material's data.

    The format of the text file is as follows:
        * The items in a line are separated by one or more spaces or tabs.
        * The first line is a heading line. It contains the word "Compound" \
        followed by zero or more assay names.
        * Subsequent lines contain a compound formula and phase, followed by \
        a mass fraction for each assay.
        * The list of compounds and mass fractions can be ended off with a \
        "#" character. This indicates that custom material properties follow \
        below in the lines below the hash.
        * If a custom material property is defined, a value must be \
        provided for each assay name. A price custom property is used as an \
        example below.

    The following is an example of a material text file::

        Compound   IlmeniteA  IlmeniteB  IlmeniteC
        Al2O3[S1]  0.01160    0.01550    0.00941
        CaO[S]     0.00022    0.00001    0.00017
        Cr2O3[S]   0.00008    0.00022    0.00011
        Fe2O3[S1]  0.20200    0.47300    0.49674
        Fe3O4[S1]  0.00000    0.00000    0.00000
        FeO[S1]    0.27900    0.19100    0.00000
        K2O[S]     0.00004    0.00001    0.00005
        MgO[S]     0.01040    0.00580    0.01090
        MnO[S]     0.00540    0.00480    0.00525
        Na2O[S1]   0.00007    0.00005    0.00031
        P4O10[S]   0.00001    0.00032    0.00015
        SiO2[S1]   0.00850    0.00490    0.01744
        TiO2[S1]   0.47700    0.29400    0.45949
        V2O5[S]    0.00360    0.00800    0.00000
        #
        Price[USD/kg]  1.2        1.3        1.1
    """

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, file_path):
        # Initialise the material's properties.
        self.name = name
        """The material's name."""

        self.compounds = list()
        """The material's list of chemical compounds."""

        # Read the material's data from the file and prepare it for use.
        f = open(file_path, "r")
        lines = f.readlines()
        f.close()
        lines = self._prepare_lines(lines)

        # Determine the assay names, and create a dictionary entry for each
        # assay.
        assay_names = lines[0].split(" ")
        del(assay_names[0:1])
        self.raw_assays = {}
        """A dictionary containing raw assays for this material."""
        self.converted_assays = {}
        """A dictionary containing converted assays for this material."""

        for assay_name in assay_names:
            self.raw_assays[assay_name] = numpy.array([])
            self.converted_assays[assay_name] = numpy.array([])

        # Read the compounds and assays.
        must_read_custom_properties = False
        for i in range(1, len(lines)):
            if lines[i].startswith("#"):
                must_read_custom_properties = True
                break
            strings = lines[i].split(" ")
            if len(strings) < len(assay_names) + 1:  # Not a full line.
                continue

            # Add the new compound.
            self.compounds.append(strings[0])

            # Add the mass fractions to the assays.
            for j in range(0, len(self.raw_assays)):
                assay_name = assay_names[j]
                self.raw_assays[assay_name] = numpy.append(self.raw_assays[assay_name],
                                                       float(strings[j+1]))
                self.converted_assays[assay_name] = numpy.append(self.converted_assays[assay_name],
                                                       float(strings[j+1]))

        # Read the custom properties.
        self.assay_custom_properties = dict()
        for j in range(0, len(self.raw_assays)):
            assay_name = assay_names[j]
            self.assay_custom_properties[assay_name] = dict()
        self.custom_properties = []
        if must_read_custom_properties:
            for i in range(i+1, len(lines)):
                strings = lines[i].split(" ")
                for j in range(0, len(self.raw_assays)):
                    assay_name = assay_names[j]
                    property_name = strings[0]
                    self.custom_properties.append(property_name)
                    property_dictionary = self.assay_custom_properties[assay_name]
                    property_dictionary[property_name] = float(strings[j+1])

        # Initialise the remaining properties.
        self.compound_count = len(self.compounds)
        """The number of chemical compounds in the material."""

        self.elements = self._create_element_list()

    def __str__(self):
        if len(self.raw_assays) > 0:
            line_length = 20 + (3 + 14) * len(self.raw_assays) - 2
        else:
            line_length = 20 + len(self.name)
        result = "=" * line_length + "\n"
        result += "Material\n"
        result += "=" * line_length + "\n"
        result += "Name".ljust(20) + self.name + "\n"

        # Create the header line of the table.
        result += "-" * line_length + "\n"
        if len(self.raw_assays) > 0:
            result = result + "Composition Details (mass fractions)\n"
        result += "Compound".ljust(20)
        assay_names = sorted(self.raw_assays.keys())
        pad = ''
        for assay_name in assay_names:
            result += pad + assay_name[0:15].ljust(15)
            pad = '  '
        result += "\n"
        result += "-" * line_length + "\n"

        # Create the content lines of the table.
        for compound in sorted(self.compounds):
            result = result + compound.ljust(20)
            compound_index = self.get_compound_index(compound)
            pad = ''
            for assay_name in assay_names:
                result += pad + '{:.8e}'.format(self.raw_assays[assay_name][compound_index]).rjust(15)
                pad = '  '
            result = result + "\n"

        # Write the total.
        result += "Total".ljust(20)
        pad = ''
        for assay_name in assay_names:
            total = 0.0
            for compound in sorted(self.compounds):
                compound_index = self.get_compound_index(compound)
                total += self.raw_assays[assay_name][compound_index]
            result += pad + '{:.8e}'.format(total).rjust(15)
            pad = '  '
        result = result + "\n"

        # Write the custom properties.
        if len(self.custom_properties) > 0:
            result += "-" * line_length + "\n"
            result += "Custom Properties:\n"
            result += "-" * line_length + "\n"
            first_assay = list(sorted(self.assay_custom_properties.keys()))[0]
            properties = list(sorted(self.assay_custom_properties[first_assay].keys()))
            for prop in properties:
                pad = ''
                result += prop.ljust(20)
                for assay_name in assay_names:
                    result += pad + "{:.8e}".format(self.assay_custom_properties[assay_name][prop]).rjust(15)
                    pad = '  '
                result += "\n"
        result += "=" * line_length + "\n"

        return result

    # -------------------------------------------------------------------------
    # Private methods.
    # -------------------------------------------------------------------------
    def _prepare_lines(self, lines):
        """Prepare the lines read from the text file before starting to
        process it.

        :param lines: The lines to prepare."""

        result = list()

        for line in lines:
            # Remove all whitespace from the start and end of the line.
            line = line.strip()

            # Replace all tabs with spaces.
            line = line.replace("\t", " ")

            # Replace all repeating spaces with a single space.
            while line.find("  ") > -1:
                line = line.replace("  ", " ")

            result.append(line)

        return result

    def _create_element_list(self):
        """Extract an alphabetically sorted list of elements from the
        material's compounds.

        :returns: Alphabetically sorted list of elements."""

        element_set = stoich.elements(self.compounds)
        return sorted(list(element_set))

    # -------------------------------------------------------------------------
    # Public methods.
    # -------------------------------------------------------------------------
    def get_compound_index(self, compound):
        """Determine the specified compound's index.

        :param compound: Formula and phase of a compound, e.g. "Fe2O3[S1]".

        :returns: Compound index."""

        return self.compounds.index(compound)

    def create_empty_assay(self):
        """Create an empty array to store an assay.

        The array's length will be equal to the number of compounds in the
        material.

        :returns: Empty assay array.
        """

        return numpy.zeros(self.compound_count)

    def add_assay(self, name, assay):
        """Add an assay to the material.

        :param name:  Assay name.
        :param assay: Numpy array containing the compound mass fractions for \
        the assay. The sequence of the assay's elements must correspond to \
        the sequence of the material's compounds.
        """

        if not type(assay) is numpy.ndarray:
            raise Exception("Invalid assay. It must be a numpy array.")
        elif not assay.shape == (self.compound_count,):
            raise Exception("Invalid assay: It must have the same number of "
                            "elements as the material has compounds.")
        elif name in self.raw_assays.keys():
            raise Exception("Invalid assay: An assay with that name already "
                            "exists.")
        self.raw_assays[name] = assay

    def get_assay_total(self, name):
        """Calculate the total/sum of the specified assay's mass fractions.

        :param name: Assay name.

        :returns: Total mass fraction."""

        return sum(self.converted_assays[name])

    def create_package(self, assay=None, mass=0.0, P=1.0, T=25.0,
                       normalise=True):
        """Create a MaterialPackage based on the specified parameters.

        :param assay:     Name of the assay to be used to create the package.
        :param mass:      Package mass. [kg]
        :param P:         Package pressure. [atm]
        :param T:         Package temperature. [°C]
        :param normalise: Indicates whether the assay must be normalised \
        before creating the package.

        :returns: MaterialPackage object."""

        if assay is None:
            return MaterialPackage(self, self.create_empty_assay(), P, T)

        if normalise:
            assay_total = self.get_assay_total(assay)
        else:
            assay_total = 1.0

        return MaterialPackage(self, mass * self.converted_assays[assay] / assay_total,
                               P, T)


class MaterialPackage(Object):
    """Represents a quantity of material consisting of multiple chemical \
    compounds, having a specific mass, pressure, temperature and enthalpy.

    :param material:        A reference to the Material to which self belongs.
    :param compound_masses: Package compound masses. [kg]
    :param P:               Package pressure. [atm]
    :param T:               Package temperature. [°C]"""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, material, compound_masses, P=1.0, T=25.0):
        # Confirm that the parameters are OK.
        if not type(material) is Material:
            raise TypeError("Invalid material type. Must be "
                            "thermomaterial.Material")
        if not type(compound_masses) is numpy.ndarray:
            raise TypeError("Invalid compound_masses type. Must be "
                            "numpy.ndarray.")

        # Initialise the object's properties.
        self.material = material
        self._P = P
        self._T = T
        self._compound_masses = compound_masses
        if self.get_mass() > 0.0:
            self._H = self._calculate_H(T)
        else:
            self._H = 0.0

        self.custom_properties = dict()


    def __str__(self):
        result = "==================================================================\n"
        result += "MaterialPackage\n"
        result += "==================================================================\n"
        result += "Material".ljust(20) + self.material.name + "\n"
        result = result + "Mass".ljust(20) + '{:.8e}'.format(self.get_mass()).rjust(15) + " kg\n"
        result = result + "Amount".ljust(20) + '{:.8e}'.format(self.get_amount()).rjust(15) + " kmol\n"
        result = result + "Pressure".ljust(20) + '{:.8e}'.format(self.P).rjust(15) + " atm\n"
        result = result + "Temperature".ljust(20) + '{:.8e}'.format(self.T).rjust(15) + " °C\n"
        result = result + "Enthalpy".ljust(20) + '{:.8e}'.format(self.H).rjust(15) + " kWh\n"
        result += "------------------------------------------------------------------\n"
        result = result + "Compound Details\n"
        result = result + "Formula".ljust(20) + "Mass".ljust(16) + \
                          "Mass Fraction".ljust(16) + \
                          "Mole Fraction".ljust(16) + "\n"
        result += "------------------------------------------------------------------\n"
        mass = self.get_mass()
        compound_moles = self.get_compound_amounts()
        total_moles = compound_moles.sum()
        if mass > 0.0:
            for compound in self.material.compounds:
                index = self.material.get_compound_index(compound)
                result += compound.ljust(20) + \
                          '{:.8e}'.format(self._compound_masses[index])
                result += "  " + \
                          '{:.8e}'.format(self._compound_masses[index] / mass)
                result += "  " + \
                          '{:.8e}'.format(compound_moles[index] / total_moles)
                result += "\n"
        else:
            for compound in self.material.compounds:
                index = self.material.get_compound_index(compound)
                result += compound.ljust(20) + '{:.8e}'.format(0.0)
                result += "  " + '{:.8e}'.format(0.0)
                result += "  " + '{:.8e}'.format(0.0)
                result += "\n"

        # Write the custom properties.
        if len(self.custom_properties) > 0:
            result += "------------------------------------------------------------------\n"
            result += "Custom Properties:\n"
            result += "------------------------------------------------------------------\n"
            properties = list(sorted(self.custom_properties.keys()))
            for prop in properties:
                result += prop.ljust(20)
                result += "{:.8e}".format(self.custom_properties[prop])
                result += "\n"

        result += "==================================================================\n"
        return result


    # -------------------------------------------------------------------------
    # Operators.
    # -------------------------------------------------------------------------
    def __add__(self, other):
        """Addition operator (+).

        Add this package (self) and 'other' together, return the result as a
        new package, and leave self unchanged.

        :param other: Can can be one of the following:
                 1. MaterialPackage
                    'other' is added to self to create a new package.
                 2. tuple: (compound, mass)
                    The specified mass of the specified compound is added to \
                    self, assuming the added material has the same \
                    temperature as self.
                 3. tuple: (compound, mass, temperature)
                    The specified mass of the specified compound at the \
                    specified temperature is added to self.

        :returns: A new Material package that is the sum of self and 'other'.
        """

        # Add another package.
        if type(other) is MaterialPackage:
            if self.material == other.material:  # Packages of same material.
                result = MaterialPackage(self.material,
                                         self._compound_masses +
                                         other._compound_masses)
                result._set_H(self._H + other._H)
                result._P = self._P
                return result
            else:  # Packages of different materials.
                H = self._get_H() + other._get_H()
                result = self.clone()
                for compound in other.material.compounds:
                    if compound not in self.material.compounds:
                        raise Exception("Packages of '" + other.material.name +
                                        "' cannot be added to packages of '" +
                                        self.material.name +
                                        "'. The compound '" + compound +
                                        "' was not found in '" +
                                        self.material.name + "'.")
                    result = result + (compound,
                                       other.get_compound_mass(compound))
                result._set_H(H)
                return result

        # Add the specified mass of the specified compound.
        elif self._is_compound_mass_tuple(other):
            # Added material varialbes.
            compound = other[0]
            index = self.material.get_compound_index(compound)
            mass = other[1]
            enthalpy = thermo.H(compound, self._T, mass)

            # Create the result package.
            result = self.clone()
            result._compound_masses[index] = result._compound_masses[index] + \
                                             mass
            result._H = result._H + enthalpy
            result._P = self._P
            return result

        # Add the specified mass of 'compound' at the specified temperature.
        elif self._is_compound_mass_temperature_tuple(other):
            # Added material varialbes.
            compound = other[0]
            index = self.material.get_compound_index(compound)
            mass = other[1]
            temperature = other[2]
            enthalpy = thermo.H(compound, temperature, mass)

            # Create the result package.
            result = self * 1.0
            result._compound_masses[index] = result._compound_masses[index] + \
                                             mass
            result._set_H(self._H + enthalpy)
            result._P = self._P
            return result

        # If not one of the above, it must be an invalid argument.
        else:
            raise TypeError("Invalid addition argument.")

    def __mul__(self, scalar):
        """The multiplication operator (*).

        Create a new package by multiplying self with scalar.

        :param scalar: The result is a new package with its content equal to \
        self multiplied by a scalar, leaving self unchanged.

        :returns: New MaterialPackage object."""

        # Multiply with a scalar floating point number.
        if type(scalar) is float or type(scalar) is numpy.float64 or \
           type(scalar) is numpy.float32:
            if scalar < 0.0:
                raise Exception("Invalid multiplication operation. Cannot "
                                "multiply package with negative number.")
            result = MaterialPackage(self.material, self._compound_masses *
                                     scalar, self._P, self._T)
            return result

        # If not one of the above, it must be an invalid argument.
        else:
            raise TypeError("Invalid multiplication argument.")


    # -------------------------------------------------------------------------
    # Private methods.
    # -------------------------------------------------------------------------
    def _calculate_H(self, T):
        """Calculate the enthalpy of the package at the specified temperature.

        :param T: Temperature. [°C]

        :returns: Enthalpy. [kWh]"""

        H = 0.0
        for compound in self.material.compounds:
            index = self.material.get_compound_index(compound)
            dH =  thermo.H(compound, T, self._compound_masses[index])
            H = H + dH
        return H

    def _calculate_T(self, H):
        """Calculate the temperature of the package given the specified
        enthalpy using a secant algorithm.

        :param H: Enthalpy. [kWh]

        :returns: Temperature. [°C]"""

        # Create the initial guesses for temperature.
        x = list()
        x.append(self._T)
        x.append(self._T + 10.0)

        # Evaluate the enthalpy for the initial guesses.
        y = list()
        y.append(self._calculate_H(x[0]) - H)
        y.append(self._calculate_H(x[1]) - H)

        # Solve for temperature.
        for i in range(2, 50):
            x.append(x[i-1] - y[i-1]*((x[i-1] - x[i-2])/(y[i-1] - y[i-2])))
            y.append(self._calculate_H(x[i]) - H)
            if abs(y[i-1]) < 1.0e-5:
                break

        return x[len(x) - 1]

    def _is_compound_mass_tuple(self, value):
        """Determines whether value is a tuple of the format
        (compound(str), mass(float)).

        :param value: The value to be tested.

        :returns: True or False"""

        if not type(value) is tuple:
            return False
        elif not len(value) == 2:
            return False
        elif not type(value[0]) is str:
            return False
        elif not type(value[1]) is float and \
             not type(value[1]) is numpy.float64 and \
             not type(value[1]) is numpy.float32:
            return False
        else:
            return True

    def _is_compound_mass_temperature_tuple(self, value):
        """Determines whether value is a tuple of the format
        (compound(str), mass(float), temperature(float)).

        :param value: The value to be tested.

        :returns: True or False"""

        if not type(value) is tuple:
            return False
        elif not len(value) == 3:
            return False
        elif not type(value[0]) is str:
            return False
        elif not type(value[1]) is float and \
             not type(value[1]) is numpy.float64 and \
             not type(value[1]) is numpy.float32:
            return False
        elif not type(value[1]) is float and \
             not type(value[1]) is numpy.float64 and \
             not type(value[1]) is numpy.float32:
            return False
        else:
            return True

    def _set_H(self, H):
        """Set the enthalpy of the package to the specified value, and
        recalculate it's temperature.

        :param H: The new enthalpy value. [kWh]"""

        self._H = H
        self._T = self._calculate_T(H)

    def _get_H(self):
        """Determine the enthalpy of the package.

        :returns: Enthalpy. [kWh]"""

        return self._H

    def _set_T(self, T):
        """Set the temperature of the package to the specified value, and
        recalculate it's enthalpy.

        :param T: Temperature. [°C]"""

        self._T = T
        self._H = self._calculate_H(T)

    def _get_T(self):
        """Determine the temperature of of the package.

        :returns: Temperature. [°C]"""

        return self._T

    def _set_P(self, P):
        """Set the pressure of the package to the specified value.

        :param P: Pressure. [atm]"""

        self._P = P

    def _get_P(self):
        """Determine the pressure of the package.

        :returns: Pressure. [atm]"""

        return self._P

    # -------------------------------------------------------------------------
    # Public methods.
    # -------------------------------------------------------------------------
    def clone(self):
        """Create a complete copy of the package.

        :returns: A new MaterialPackage object."""

        result = copy.copy(self)
        result._compound_masses = copy.deepcopy(self._compound_masses)
        return result

    def clear(self):
        """Clear the package."""

        self._compound_masses = self._compound_masses * 0.0
        self._P = 1.0
        self._T = 25.0
        self._H = 0.0

    def get_assay(self):
        """Determine the assay of the package.

        :returns: Array of mass fractions."""

        return self._compound_masses / self._compound_masses.sum()

    def get_mass(self):
        """Determine the mass of the package.

        :returns: Mass. [kg]"""

        return self._compound_masses.sum()

    def get_compound_mass(self, compound):
        """Determine the mass of the specified compound in the package.

        :param compound: Formula and phase of a compound, e.g. "Fe2O3[S1]".

        :returns: Mass. [kg]"""

        if compound in self.material.compounds:
            return self._compound_masses[self.material.get_compound_index(compound)]
        else:
            return 0.0

    def get_compound_amounts(self):
        """Determine the mole amounts of all the compounds.

        :returns: List of amounts. [mol]"""

        result = self._compound_masses * 1.0
        for compound in self.material.compounds:
            index = self.material.get_compound_index(compound)
            result[index] = stoich.amount(compound, result[index])
        return result

    def get_compound_amount(self, compound):
        """Determine the mole amount of the specified compound.

        :returns: Amount. [mol]"""

        index = self.material.get_compound_index(compound)
        result = self._compound_masses[index]
        result = stoich.amount(compound, result)
        return result

    def get_amount(self):
        """Determine the sum of mole amounts of all the compounds.

        :returns: Amount. [mol]
        """

        result = 0.0
        for compound in self.material.compounds:
            result += self.get_compound_amount(compound)
        return result

    H = property(_get_H, _set_H, None, "Enthalpy. [kWh]")
    T = property(_get_T, _set_T, None, "Temperature. [°C]")
    P = property(_get_P, _set_P, None, "Pressure. [atm]")
    mass = property(get_mass, None, None, "[kg]")
    amount = property(get_amount, None, None, "[kmol]")

    def get_element_masses(self, elements = None):
        """Determine the masses of elements in the package.

        :returns: Array of element masses. [kg]"""

        if elements == None:
            elements = self.material.elements
        result = numpy.zeros(len(elements))
        for compound in self.material.compounds:
            result = result + self.get_compound_mass(compound) * stoich.element_mass_fractions(compound, elements)
        return result

    def get_element_mass_dictionary(self):
        """Determine the masses of elements in the package and return as a
        dictionary.

        :returns: Dictionary of element symbols and masses. [kg]"""

        element_symbols = self.material.elements
        element_masses = self.get_element_masses()
        result = dict()
        for s, m in zip(element_symbols, element_masses):
            result[s] = m
        return result

    def get_element_mass(self, element):
        """Determine the mass of the specified elements in the package.

        :returns: Masses. [kg]"""

        result = numpy.zeros(1)
        for compound in self.material.compounds:
            result = result + self.get_compound_mass(compound) * stoich.element_mass_fractions(compound, [element])
        return result[0]

    def extract(self, other):
        """Extract some material from this package.
        Extract 'other' from this package, modifying this package and
        returning the extracted material as a new package.

        :param other: Can be one of the following:

                 1. float
                    A mass equal to other is extracted from self. Self is
                    reduced by other and the extracted package is returned as
                    a new package.
                 2. tuple: (compound, mass)
                    The other tuple specifies the mass of a compound to be
                    extracted. It is extracted from self and the extracted
                    mass is returned as a new package.
                 3. string
                    The 'other' string specifies the compound to be extracted.
                    All of the mass of that compound will be removed from self
                    and a new package created with it.
                 4. Material
                    The 'other' material specifies the list of compounds to
                    extract.

        :returns: New MaterialPackage object."""

        # Extract the specified mass.
        if type(other) is float or \
           type(other) is numpy.float64 or \
           type(other) is numpy.float32:
            return self._extract_mass(other)

        # Extract the specified mass of the specified compound.
        elif self._is_compound_mass_tuple(other):
            return self._extract_compound_mass(other[0], other[1])

        # Extract all of the specified compound.
        elif type(other) is str:
            return self._extract_compound(other)

        # TODO: Test
        # Extract all of the compounds of the specified material.
        elif type(other) is Material:
            return self._extract_material(other)

        # If not one of the above, it must be an invalid argument.
        else:
            raise TypeError("Invalid extraction argument.")

    def _extract_mass(self, mass):
        if mass > self.get_mass():
            raise Exception("Invalid extraction operation. Cannot extract a mass larger than the package's mass.")
        fraction_to_subtract = mass / self.get_mass()
        result = MaterialPackage(self.material, self._compound_masses * fraction_to_subtract, self._P, self._T)

        self._compound_masses = self._compound_masses * (1.0 - fraction_to_subtract)
        self.T = self.T

        return result

    def _extract_compound(self, compound):
        result = self.material.create_package()

        if compound not in self.material.compounds:
            return result

        index = self.material.get_compound_index(compound)
        result._compound_masses[index] = self._compound_masses[index]
        result.T = self.T

        self._compound_masses[index] = 0.0
        self.T = self.T

        return result

    def _extract_compound_mass(self, compound, mass):
        if compound not in self.material.compounds:
            return self.material.create_package()

        index = self.material.get_compound_index(compound)
        if mass > self._compound_masses[index]:
            raise Exception("Invalid extraction operation. Cannot extract a compound mass larger than what the package contains.")
        self._compound_masses[index] = self._compound_masses[index] - mass
        self.T = self.T

        result = self.material.create_package(P=self._P, T=self._T)
        result += (compound, mass)

        return result

    def _extract_material(self, material):
        result = material.create_package()
        for compound in material.compounds:
            mass = self.get_compound_mass(compound)
            result += (compound, mass)
            self.extract(compound)
        result.T = self.T

        return result


class AssayConversionAction(object):
    def run(self, material):
        pass


class Normalise(AssayConversionAction):
    def run(self, material):
        pass


def _get_default_data_path():
    module_path = os.path.dirname(sys.modules[__name__].__file__)
    data_path = os.path.join(module_path, r"../data")
    data_path = os.path.abspath(data_path)
    return data_path

DEFAULT_DATA_PATH = _get_default_data_path()



"""auxi.modeling.financial.des Package

The purpose of this package is to provide common classes and functions for the the auxi.modeling.financial.des (double entry system) namespace"""


__version__ = "0.2.0"



# -*- coding: utf-8 -*-
"""
This module provides testing code for the
auxi.modeling.financial.tax.capitalgainsrule module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
from auxi.modeling.financial.tax.capitalgainsrule import CapitalGainsRule

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class TestAllFunctions(unittest.TestCase):
    """
      Tester for the auxi.modeling.financial.tax.capitalgainsrule class.
    """

    def setUp(self):
        self.object = CapitalGainsRule("NameA",
                                       description="DescriptionA",
                                       percentage=1.14)

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.percentage, 1.14)


# =============================================================================
# Display documentation and run tests.
# =============================================================================
# os.system("cls")

# help(SalesRule)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Created on Mon Jan 18 05:34:00 2016

@author: johanz
"""

from auxi.core.object import Object


class NamedObject(Object):
    """
    Base class for all auxi classes containing a name and description.
    """

    def __init__(self, name, description):
        self.name = name
        self.description = description

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value



# -*- coding: utf-8 -*-
"""
This module provides a transaction template class that serves as a template
for how a transaction is to be created.\n

@name: transaction template
@author: Ex Mente Technologies (Pty) Ltd
"""

from auxi.core.namedobject import NamedObject


__version__ = "0.2.0"


class TransactionTemplate(NamedObject):
    """Represents a template for how a transaction is to be created."""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, description=None,
                 dt_account=None, cr_account=None):
        """Initialise the object."""
        super().__init__(name, description)
        self.dt_account = dt_account
        self.cr_account = cr_account



# -*- coding: utf-8 -*-
"""
This module provides testing code for the
auxi.modeling.financial.des.generalledger module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
from datetime import datetime
from auxi.modeling.financial.des.generalledgeraccount import AccountType
from auxi.modeling.financial.des.generalledgerstructure import GeneralLedgerStructure
from auxi.modeling.financial.des.generalledger import GeneralLedger

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class TestAllFunctions(unittest.TestCase):
    """
      Tester for the auxi.modeling.financial.des.generalledger class.
    """

    def setUp(self):
        self.structure = GeneralLedgerStructure("NameA",
                                                description="DescriptionA")
        self.structure.create_account("TestA",
                                      description="TestA_Desc",
                                      number="010",
                                      account_type=AccountType.equity)
        self.object = GeneralLedger("NameA",
                                    self.structure,
                                    description="DescriptionA")

    def test_constructor(self):
        self.assertEqual(self.object.name, "NameA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.structure, self.object.structure)

    def test_create_transaction(self):
        new_tx = self.object.create_transaction("TestA",
                                                description="TestA_Desc",
                                                tx_datetime=datetime(2016, 2, 1),
                                                dt_account="Bank",
                                                cr_account="Sales",
                                                source="Peanut Sales",
                                                amount=20.00)

        self.assertEqual(new_tx.name, self.object.transactions[0].name)
        self.assertEqual(new_tx.tx_datetime, self.object.transactions[0].tx_datetime)
        self.assertEqual(new_tx.dt_account, self.object.transactions[0].dt_account)
        self.assertEqual(new_tx.cr_account, self.object.transactions[0].cr_account)
        self.assertEqual(new_tx.source, self.object.transactions[0].source)
        self.assertEqual(new_tx.amount, self.object.transactions[0].amount)


# =============================================================================
# Display documentation and run tests.
# =============================================================================
# os.system("cls")

# help(TransactionTemplate)

if __name__ == '__main__':
    unittest.main()



# USAGE: you probably want 'setup.py install' - but execute 'setup.py --help'
# for all the details.

# from setuptools import setup, Extension, Command

from distutils.core import setup

# build the distribution
setup(name="auxi",
      version="0.2.0rc3",
      description="A toolkit to help metallurgical process engineers to rapidly do day-to-day calculations.",
      long_description=open('README.txt').read(),
      author="Ex Mente (Pty) Ltd",
      author_email="dev@ex-mente.co.za",
      maintainer="Ex Mente (Pty) Ltd",
      maintainer_email="dev@ex-mente.co.za",
      url="https://github.com/Ex-Mente/auxi.0",
      download_url="https://pypi.python.org/pypi/auxi",
      license='LICENSE.txt',
      keywords="metallurgy,chemistry,modelling,simulation,thermochemistry,engineering,mass balance,energy balance",
      platforms="Ubunutu 14.04,Ubuntu 15.04,Windows 7 (For the stoichiometry and thermochemistry tools only.)",
      package_dir={'auxi': 'src'},
      packages=["auxi",
                "auxi.core",
                "auxi.modeling",
                "auxi.modeling.process",
                "auxi.modeling.process.materials",
                "auxi.modeling.process.materials.chemistry",
                "auxi.modeling.process.materials.psd",
                "auxi.modeling.process.materials.thermochemistry",
                "auxi.modeling.financial",
                "auxi.modeling.financial.des",
                "auxi.modeling.financial.tax",
                "auxi.modeling.business",
                "auxi.tools",
                "auxi.tools.chemistry"],
      package_data={'auxi.tools.chemistry': [r'data/thermo/*.txt'],
                    'auxi.modeling.process.materials': [r'data/*.txt'],
                    'auxi': [r'*.txt']}
      )



# -*- coding: utf-8 -*-
"""
This module provides a transaction class that stores information on a
transaction between two general ledger accounts.\n

@name: transaction
@author: Ex Mente Technologies (Pty) Ltd
"""

from datetime import datetime
from auxi.core.namedobject import NamedObject


__version__ = "0.2.0"


class Transaction(NamedObject):
    """Represents a financial transaction between
       two general ledger accounts.
    """

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, description=None,
                 tx_datetime=datetime.min,
                 dt_account=None, cr_account=None,
                 source=None, amount=0.00,
                 is_closing_dt_account=False,
                 is_closing_cr_account=False):
        """Initialise the object."""
        super().__init__(name, description)
        self.tx_datetime = tx_datetime
        self.dt_account = dt_account
        self.cr_account = cr_account
        self.source = source
        self.amount = amount
        self.is_closing_dt_account = is_closing_dt_account
        self.is_closing_cr_account = is_closing_cr_account



# -*- coding: utf-8 -*-
"""
This module provides material and material package classes that can do chemical calculations.\n

@name: chemistry
@author: Ex Mente Technologies (Pty) Ltd
"""

import os
import sys
import numpy
import copy
from auxi.core.object import Object
from auxi.core.namedobject import NamedObject
from auxi.tools.chemistry import stoichiometry as stoich

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class Material(NamedObject):
    """Represents a material consisting of multiple chemical compounds.\n
    Properties defined here:\n
    name           : The material's name.
    compounds      : The material's list of chemical compounds.
    compound_count : The number of chemical compounds in the material.
    assays         : A dictionary containing assays for this material.
    """

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, file_path):
        """Initialise the object from a text file containing compounds and assays.
        The format of the text file is as follows:
        * The lines are space separated. The values in a line are separated by one or more spaces.
        * The first line is a heading line.
        * All subsequent lines contain a compound formula and phase, followed by mass fractions.
        * The first column lists the compounds in the material.
        * All subsequent columns describe assays of the material.
        The following is an example of a material text file:
        Compound   IlmeniteA  IlmeniteB  IlmeniteC
        Al2O3      0.01160    0.01550    0.00941
        CaO        0.00022    0.00001    0.00017
        Cr2O3      0.00008    0.00022    0.00011
        Fe2O3      0.20200    0.47300    0.49674
        Fe3O4      0.00000    0.00000    0.00000
        FeO        0.27900    0.19100    0.00000
        K2O        0.00004    0.00001    0.00005
        MgO        0.01040    0.00580    0.01090
        MnO        0.00540    0.00480    0.00525
        Na2O       0.00007    0.00005    0.00031
        P4O10      0.00001    0.00032    0.00015
        SiO2       0.00850    0.00490    0.01744
        TiO2       0.47700    0.29400    0.45949
        V2O5       0.00360    0.00800    0.00000
        """

        # Initialise the material's properties.
        self.name = name
        self.compounds = list()

        # Read the material's data from the file and prepare it for use.
        f = open(file_path, "r")
        lines = f.readlines()
        f.close()
        lines = self._prepare_lines(lines)

        # Determine the assay names, and create a dictionary entry for each assay.
        assay_names = lines[0].split(" ")
        del(assay_names[0:1])
        self.assays = dict()
        for assay_name in assay_names:
            self.assays[assay_name] = numpy.array([])

        # Read the compounds and assays.
        for i in range(1, len(lines)):
            strings = lines[i].split(" ")
            if len(strings) < len(assay_names) + 1: # Not a full line.
                continue
            self.compounds.append(strings[0])       # Add the new compound.
            for j in range(0, len(self.assays)):    # Add the mass fractions to the assays.
                assay_name = assay_names[j]
                self.assays[assay_name] = numpy.append(self.assays[assay_name], float(strings[j+1]))

        # Initialise the remaining properties.
        self.compound_count = len(self.compounds)
        self.elements = self._create_element_list()


    def __str__(self):
        """Create a string representation of self."""
        result = "Material: name='" + self.name + "'\n"

        # Create the header line of the table.
        result = result + "Compound".ljust(20)
        assay_names = sorted(self.assays.keys())
        for assay_name in assay_names:
            result = result + assay_name.ljust(20)
        result = result + "\n"

        # Create the content lines of the table.
        for compound in self.compounds:
            result = result + compound.ljust(20)
            compound_index = self.get_compound_index(compound)
            for assay_name in assay_names:
                result = result + str(self.assays[assay_name][compound_index]).ljust(20)
            result = result + "\n"
        return result


    # -------------------------------------------------------------------------
    # Private methods.
    # -------------------------------------------------------------------------
    def _prepare_lines(self, lines):
        """Prepare the lines read from the text file before starting to process it."""

        result = list()
        for line in lines:
            line = line.strip()                 # Remove all whitespace characters (e.g. spaces, line breaks, etc.) from the start and end of the line.
            line = line.replace("\t", " ")      # Replace all tabs with spaces.
            while line.find("  ") > -1:         # Replace all repeating spaces with a single space.
                line = line.replace("  ", " ")
            result.append(line)
        return result


    def _create_element_list(self):
        """Extract an alphabetically sorted list of elements from the compounds of the material.\n
        return : An alphabeticall sorted list of elements."""

        element_set = stoich.elements(self.compounds)
        return sorted(list(element_set))


    # -------------------------------------------------------------------------
    # Public methods.
    # -------------------------------------------------------------------------
    def get_compound_index(self, compound):
        """Determine the index of the specified compound.\n
        compound : The formula and phase of the specified compound, e.g. Fe2O3[S1].\n
        return   : The index of the specified compound.
        """

        return self.compounds.index(compound)


    def create_empty_assay(self):
        """Create an empty array to store an assay. The array's length will be equal to the number of compounds in the material.\n
        return : A floating point array.
        """

        return numpy.zeros(self.compound_count)


    def add_assay(self, name, assay):
        """Add an assay to the material.\n
        name  : The name of the new assay.
        assay : A numpy array containing the compound mass fractions for the assay. The sequence of the assay's elements must correspond to the sequence of the material's compounds.
        """

        if not type(assay) is numpy.ndarray:
            raise Exception("Invalid assay. It must be a numpy array.")
        elif not assay.shape == (self.compound_count,):
            raise Exception("Invalid assay: It must have the same number of elements as the material has compounds.")
        elif name in self.assays.keys():
            raise Exception("Invalid assay: An assay with that name already exists.")
        self.assays[name] = assay


    def get_assay_total(self, name):
        """Calculate the total of the specified assay.\n
        name   : The name of the assay.\n
        return : The total mass fraction of the specified assay.
        """

        return sum(self.assays[name])


    def create_package(self, assay = None, mass = 0.0, normalise=True):
        """Create a MaterialPackage based on the specified parameters.\n
        assay     :       The name of the assay based on which the package must be created.
        mass      : [kg]  The mass of the package.
        normalise :       Indicates whether the assay must be normalised before creating the package.\n
        return    :       The created MaterialPackage.
        """

        if assay == None:
            return MaterialPackage(self, self.create_empty_assay())

        if normalise:
            assay_total = self.get_assay_total(assay)
        else:
            assay_total = 1.0
        return MaterialPackage(self, mass * self.assays[assay] / assay_total)



class MaterialPackage(Object):

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, material, compound_masses):
        """Initialise the object.\n
        material        :       A reference to the Material to which self belongs.
        compound_masses : [kg]  The masses of the compounds in the package.
        """

        # Confirm that the parameters are OK.
        if not type(material) is Material:
            raise TypeError("Invalid material type. Must be chemistry.material.Material")
        if not type(compound_masses) is numpy.ndarray:
            raise TypeError("Invalid compound_masses type. Must be numpy.ndarray.")

        # Initialise the object's properties.
        self.material = material
        self.compound_masses = compound_masses


    def __str__(self):
        """Create a string representation of the object."""
        result = "MaterialPackage\n"
        result = result + "material".ljust(20) + self.material.name + "\n"
        result = result + "mass".ljust(20) + str(self.get_mass()) + "\n"
        result = result + "Component masses:\n"
        for compound in self.material.compounds:
            index = self.material.get_compound_index(compound)
            result = result + compound.ljust(20) + str(self.compound_masses[index]) + "\n"
        return result


    # -------------------------------------------------------------------------
    # Operators.
    # -------------------------------------------------------------------------
    def __add__(self, other):
        """Addition operator (+).
        Add self and 'other' together, return the result as a new package, and leave self unchanged.\n
        other  : Can can be one of the following:
                 1. MaterialPackage
                    'other' is added to self to create a new package.
                 2. tuple: (compound, mass)
                    The specified mass of the specified compound is added to self.
        return : A new Material package that is the sum of self and 'other'.
        """

        # Add another package.
        if type(other) is MaterialPackage:
            if self.material == other.material: # Packages of the same material.
                result =  MaterialPackage(self.material, self.compound_masses + other.compound_masses)
                return result
            else: # Packages of different materials.
                result = self.clone()
                for compound in other.material.compounds:
                    if compound not in self.material.compounds:
                        raise Exception("Packages of '" + other.material.name + "' cannot be added to packages of '" + self.material.name + "'. The compound '" + compound + "' was not found in '" + self.material.name + "'.")
                    result = result + (compound, other.get_compound_mass(compound))
                return result

        # Add the specified mass of the specified compound.
        elif self._is_compound_mass_tuple(other):
            # Added material varialbes.
            compound = other[0]
            compound_index = self.material.get_compound_index(compound)
            mass = other[1]

            # Create the result package.
            result = self.clone()
            result.compound_masses[compound_index] = result.compound_masses[compound_index] + mass
            return result

        # If not one of the above, it must be an invalid argument.
        else:
            raise TypeError("Invalid addition argument.")


    def __mul__(self, scalar):
        """The multiplication operator (*).
        Create a new package by multiplying self with other.\n
        scalar : The result is a new package with its content equal to self multiplied by a scalar, leaving self unchanged.\n
        result : A new MaterialPackage equal to self package multiplied by other.
        """

        # Multiply with a scalar floating point number.
        if type(scalar) is float or type(scalar) is numpy.float64 or type(scalar) is numpy.float32:
            if scalar < 0.0:
                raise Exception("Invalid multiplication operation. Cannot multiply package with negative number.")
            result = MaterialPackage(self.material, self.compound_masses * scalar)
            return result

        # If not one of the above, it must be an invalid argument.
        else:
            raise TypeError("Invalid multiplication argument.")


    # -------------------------------------------------------------------------
    # Private methods.
    # -------------------------------------------------------------------------
    def _is_compound_mass_tuple(self, value):
        """Determines whether value is a tuple of the format (compound(str), mass(float))."""

        if not type(value) is tuple:
            return False
        elif not len(value) == 2:
            return False
        elif not type(value[0]) is str:
            return False
        elif not type(value[1]) is float and not type(value[1]) is numpy.float64 and not type(value[1]) is numpy.float32:
            return False
        else:
            return True


    # -------------------------------------------------------------------------
    # Public methods.
    # -------------------------------------------------------------------------
    def clone(self):
        """Create a complete copy of self.\n
        return : A MaterialPackage that is identical to self."""

        result = copy.copy(self)
        result.compound_masses = copy.deepcopy(self.compound_masses)
        return result


    # TODO: document
    # TODO: test
    def clear(self):
        self.compound_masses = self.compound_masses * 0.0

    def get_assay(self):
        """Determine the assay of self.\n
        return : [mass fractions] An array containing the assay of self.
        """

        return self.compound_masses / self.compound_masses.sum()


    def get_mass(self):
        """Determine the mass of self.\n
        return : [kg] The mass of self.
        """

        return self.compound_masses.sum()


    def get_compound_mass(self, compound):
        """Determine the mass of the specified compound in self.\n
        compound :      The formula and phase of the compound, e.g. Fe2O3[S1]\n
        return   : [kg] The mass of the compound in self.
        """

        return self.compound_masses[self.material.get_compound_index(compound)]


    # TODO: Test
    def get_compound_mass_fraction(self, compound):
        """Determine the mass fraction of the specified compound in self.\n
        compound : The formula and phase of the compound, e.g. Fe2O3[S1]\n
        return   : The mass fraction of the compound in self.
        """

        return self.get_compound_mass(compound) / self.get_mass()


    def get_element_masses(self):
        """Determine the masses of elements in the package.\n
        return : [kg] An array of element masses. The sequence of the elements in the result corresponds with the sequence of elements in the element list of the material.
        """

        result = numpy.zeros(len(self.material.elements))
        for compound in self.material.compounds:
            result = result + self.get_compound_mass(compound) * stoich.element_mass_fractions(compound, self.material.elements)
        return result


    def get_element_mass_dictionary(self):
        """Determine the masses of elements in the package and return as a dictionary.\n
        return : [kg] A dictionary of element symbols and masses.
        """

        element_symbols = self.material.elements
        element_masses = self.get_element_masses()
        result = dict()
        for s, m in zip(element_symbols, element_masses):
            result[s] = m
        return result


    def get_element_mass(self, element):
        """Determine the masses of elements in the package.\n
        return : [kg] An array of element masses. The sequence of the elements in the result corresponds with the sequence of elements in the element list of the material.
        """

        result = numpy.zeros(1)
        for compound in self.material.compounds:
            result = result + self.get_compound_mass(compound) * stoich.element_mass_fractions(compound, [element])
        return result[0]

    def extract(self, other):
        """Extract some material from self.
        Extract 'other' from self, modifying self and returning the extracted material as a new package.\n
        other  : Can be one of the following:
                 1. float
                    A mass equal to other is extracted from self. Self is reduced by other and the extracted package is returned as a new package.
                 2. tuple: (compound, mass)
                    The other tuple specifies the mass of a compound to be extracted. It is extracted from self and the extracted mass is returned as a new package.
                 3. string
                    The 'other' string specifies the compound to be extracted. All of the mass of that compound will be removed from self and a new package created with it.\n
        return : A new material package containing the material that was extracted from self.
        """

        # Extract the specified mass.
        if type(other) is float or type(other) is numpy.float64 or type(other) is numpy.float32:
            if other > self.get_mass():
                raise Exception("Invalid extraction operation. Cannot extract a mass larger than the package's mass.")
            fraction_to_subtract = other / self.get_mass()
            result = MaterialPackage(self.material, self.compound_masses * fraction_to_subtract)
            self.compound_masses = self.compound_masses * (1.0 - fraction_to_subtract)
            return result

        # Extract the specified mass of the specified compound.
        elif self._is_compound_mass_tuple(other):
            index = self.material.get_compound_index(other[0])
            if other[1] > self.compound_masses[index]:
                raise Exception("Invalid extraction operation. Cannot extract a compound mass larger than what the package contains.")
            self.compound_masses[index] = self.compound_masses[index] - other[1]
            resultarray = self.compound_masses*0.0
            resultarray[index] = other[1]
            result = MaterialPackage(self.material, resultarray)
            return result

        # Extract all of the specified compound.
        elif type(other) is str:
            index = self.material.get_compound_index(other)
            result = self * 0.0
            result.compound_masses[index] = self.compound_masses[index]
            self.compound_masses[index] = 0.0
            return result

        # If not one of the above, it must be an invalid argument.
        else:
            raise TypeError("Invalid extraction argument.")


    # TODO: Test
    # TODO: Document
    def add_to(self, other):
        # Add another package.
        if type(other) is MaterialPackage:
            if self.material == other.material: # Packages of the same material.
                self.compound_masses = self.compound_masses + other.compound_masses
            else: # Packages of different materials.
                for compound in other.material.compounds:
                    if compound not in self.material.compounds:
                        raise Exception("Packages of '" + other.material.name + "' cannot be added to packages of '" + self.material.name + "'. The compound '" + compound + "' was not found in '" + self.material.name + "'.")
                    self.add_to((compound, other.get_compound_mass(compound)))

        # Add the specified mass of the specified compound.
        elif self._is_compound_mass_tuple(other):
            # Added material varialbes.
            compound = other[0]
            compound_index = self.material.get_compound_index(compound)
            mass = other[1]

            # Create the result package.
            self.compound_masses[compound_index] = self.compound_masses[compound_index] + mass

        # If not one of the above, it must be an invalid argument.
        else:
            raise TypeError("Invalid addition argument.")


def _get_default_data_path():
    module_path = os.path.dirname(sys.modules[__name__].__file__)
    data_path = os.path.join(module_path, r"../data")
    data_path = os.path.abspath(data_path)
    return data_path

DEFAULT_DATA_PATH = _get_default_data_path()
